import pytest

from globus_sdk.services.gcs import UnpackingGCSResponse


def test_unpacking_response_with_callback(make_response):
    def identify_desired_data(d):
        return "foo" in d

    base_resp = make_response(
        json_body={
            "data": [
                {"x": 1, "y": 2},
                {"x": 2, "y": 3, "foo": "bar"},
            ]
        }
    )

    resp = UnpackingGCSResponse(base_resp, identify_desired_data)
    assert resp["x"] == 2
    assert resp["y"] == 3
    assert resp["foo"] == "bar"


@pytest.mark.parametrize("datatype", ["foo#1.0.1", "foo#1", "foo#0.0.1", "foo#2.0.0.1"])
def test_unpacking_response_matches_datatype(make_response, datatype):
    base_resp = make_response(json_body={"data": [{"x": 1, "DATA_TYPE": datatype}]})
    resp = UnpackingGCSResponse(base_resp, "foo")
    assert "x" in resp
    # using some other spec will not match
    resp_bar = UnpackingGCSResponse(base_resp, "bar")
    assert "x" not in resp_bar


def test_unpacking_response_invalid_spec(make_response):
    base_resp = make_response(json_body={"data": [{"x": 1, "DATA_TYPE": "foo#1.0.0"}]})
    with pytest.raises(ValueError):
        UnpackingGCSResponse(base_resp, "foo 1.0")


def test_unpacking_response_invalid_datatype(make_response):
    # we'll never return a match if the DATA_TYPE doesn't appear to be valid
    base_resp = make_response(json_body={"data": [{"x": 1, "DATA_TYPE": "foo"}]})
    resp = UnpackingGCSResponse(base_resp, "foo")
    assert "x" not in resp



import time
from unittest import mock

import pytest

from globus_sdk.globus_app.authorizer_factory import (
    AccessTokenAuthorizerFactory,
    ClientCredentialsAuthorizerFactory,
    RefreshTokenAuthorizerFactory,
)
from globus_sdk.token_storage import (
    HasRefreshTokensValidator,
    MemoryTokenStorage,
    NotExpiredValidator,
)
from globus_sdk.token_storage.validating_token_storage import (
    ExpiredTokenError,
    MissingTokenError,
    ValidatingTokenStorage,
)


def make_mock_token_response(token_number=1):
    ret = mock.Mock()
    ret.by_resource_server = {
        "rs1": {
            "resource_server": "rs1",
            "scope": "rs1:all",
            "access_token": f"rs1_access_token_{token_number}",
            "refresh_token": f"rs1_refresh_token_{token_number}",
            "expires_at_seconds": int(time.time()) + 3600,
            "token_type": "Bearer",
        }
    }
    ret.decode_id_token.return_value = {"sub": "dummy_id"}
    return ret


def _make_mem_token_storage(validators=()):
    return ValidatingTokenStorage(MemoryTokenStorage(), validators=validators)


def test_access_token_authorizer_factory():
    initial_response = make_mock_token_response()
    mock_token_storage = _make_mem_token_storage()
    mock_token_storage.store_token_response(initial_response)
    factory = AccessTokenAuthorizerFactory(token_storage=mock_token_storage)

    # cache is initially empty
    assert factory._authorizer_cache == {}

    # calling get_authorizer once creates a new authorizer from underlying storage
    authorizer = factory.get_authorizer("rs1")
    assert authorizer.get_authorization_header() == "Bearer rs1_access_token_1"

    # calling get_authorizer again gets the same cached authorizer
    authorizer2 = factory.get_authorizer("rs1")
    assert authorizer is authorizer2

    # calling store_token_response_and_clear_cache then get gets a new authorizer
    new_data = make_mock_token_response(token_number=2)
    factory.store_token_response_and_clear_cache(new_data)
    assert factory._authorizer_cache == {}
    authorizer = factory.get_authorizer("rs1")
    assert authorizer.get_authorization_header() == "Bearer rs1_access_token_2"


def test_access_token_authorizer_factory_no_tokens():
    initial_response = make_mock_token_response()
    mock_token_storage = _make_mem_token_storage()
    mock_token_storage.store_token_response(initial_response)
    factory = AccessTokenAuthorizerFactory(token_storage=mock_token_storage)

    with pytest.raises(MissingTokenError) as exc:
        factory.get_authorizer("rs2")
    assert str(exc.value) == "No token data for rs2"


def test_access_token_authorizer_factory_expired_access_token():
    initial_response = make_mock_token_response()
    initial_response.by_resource_server["rs1"]["expires_at_seconds"] = int(
        time.time() - 3600
    )

    mock_token_storage = _make_mem_token_storage(validators=(NotExpiredValidator(),))
    mock_token_storage.store_token_response(initial_response)
    factory = AccessTokenAuthorizerFactory(token_storage=mock_token_storage)

    with pytest.raises(ExpiredTokenError):
        factory.get_authorizer("rs1")


def test_refresh_token_authorizer_factory():
    initial_response = make_mock_token_response()
    mock_token_storage = _make_mem_token_storage(
        validators=(HasRefreshTokensValidator(),)
    )
    mock_token_storage.store_token_response(initial_response)

    refresh_data = make_mock_token_response(token_number=2)
    mock_auth_login_client = mock.Mock()
    mock_refresh = mock.Mock()
    mock_refresh.return_value = refresh_data
    mock_auth_login_client.oauth2_refresh_token = mock_refresh

    factory = RefreshTokenAuthorizerFactory(
        token_storage=mock_token_storage,
        auth_login_client=mock_auth_login_client,
    )

    # calling get authorizer creates a new authorizer with existing token data
    authorizer1 = factory.get_authorizer("rs1")
    assert authorizer1.get_authorization_header() == "Bearer rs1_access_token_1"
    assert mock_auth_login_client.oauth2_refresh_token.call_count == 0

    # standard refresh doesn't change the authorizer
    authorizer1._get_new_access_token()
    authorizer2 = factory.get_authorizer("rs1")
    assert authorizer2 is authorizer1
    assert authorizer2.get_authorization_header() == "Bearer rs1_access_token_2"
    assert mock_auth_login_client.oauth2_refresh_token.call_count == 1

    # calling store_token_response_and_clear_cache then get gets a new authorizer
    factory.store_token_response_and_clear_cache(initial_response)
    authorizer3 = factory.get_authorizer("rs1")
    assert authorizer3 is not authorizer1
    assert authorizer3.get_authorization_header() == "Bearer rs1_access_token_1"


def test_refresh_token_authorizer_factory_expired_access_token():
    initial_response = make_mock_token_response()
    initial_response.by_resource_server["rs1"]["expires_at_seconds"] = int(
        time.time() - 3600
    )

    mock_token_storage = _make_mem_token_storage(
        validators=(HasRefreshTokensValidator(),)
    )
    mock_token_storage.store_token_response(initial_response)

    refresh_data = make_mock_token_response(token_number=2)
    mock_auth_login_client = mock.Mock()
    mock_refresh = mock.Mock()
    mock_refresh.return_value = refresh_data
    mock_auth_login_client.oauth2_refresh_token = mock_refresh

    factory = RefreshTokenAuthorizerFactory(
        token_storage=mock_token_storage,
        auth_login_client=mock_auth_login_client,
    )

    # calling get_authorizer automatically causes a refresh call to get an access token
    authorizer = factory.get_authorizer("rs1")
    assert authorizer.get_authorization_header() == "Bearer rs1_access_token_2"
    assert mock_refresh.call_count == 1


def test_refresh_token_authorizer_factory_no_refresh_token():
    initial_response = make_mock_token_response()
    initial_response.by_resource_server["rs1"]["refresh_token"] = None

    mock_token_storage = _make_mem_token_storage(
        validators=(HasRefreshTokensValidator(),)
    )
    mock_token_storage.store_token_response(initial_response)

    factory = RefreshTokenAuthorizerFactory(
        token_storage=mock_token_storage,
        auth_login_client=mock.Mock(),
    )

    with pytest.raises(MissingTokenError) as exc:
        factory.get_authorizer("rs1")
    assert str(exc.value) == "No refresh_token for rs1"


def test_client_credentials_authorizer_factory():
    initial_response = make_mock_token_response()
    mock_token_storage = _make_mem_token_storage()
    mock_token_storage.store_token_response(initial_response)

    client_token_data = make_mock_token_response(token_number=2)
    mock_confidential_client = mock.Mock()
    mock_client_credentials_tokens = mock.Mock()
    mock_client_credentials_tokens.return_value = client_token_data
    mock_confidential_client.oauth2_client_credentials_tokens = (
        mock_client_credentials_tokens
    )

    factory = ClientCredentialsAuthorizerFactory(
        token_storage=mock_token_storage,
        confidential_client=mock_confidential_client,
        scope_requirements={"rs1": ["scope1"]},
    )

    # calling get_authorizer once creates a new authorizer using existing tokens
    authorizer1 = factory.get_authorizer("rs1")
    assert authorizer1.get_authorization_header() == "Bearer rs1_access_token_1"
    assert mock_confidential_client.oauth2_client_credentials_tokens.call_count == 0

    # renewing with existing tokens doesn't change the authorizer
    authorizer1._get_new_access_token()
    authorizer2 = factory.get_authorizer("rs1")
    assert authorizer2 is authorizer1
    assert authorizer2.get_authorization_header() == "Bearer rs1_access_token_2"
    assert mock_confidential_client.oauth2_client_credentials_tokens.call_count == 1

    # calling store_token_response_and_clear_cache then get gets a new authorizer
    factory.store_token_response_and_clear_cache(initial_response)
    authorizer3 = factory.get_authorizer("rs1")
    assert authorizer3 is not authorizer1
    assert authorizer3.get_authorization_header() == "Bearer rs1_access_token_1"


def test_client_credentials_authorizer_factory_no_tokens():
    mock_token_storage = _make_mem_token_storage()

    client_token_data = make_mock_token_response()
    mock_confidential_client = mock.Mock()
    mock_client_credentials_tokens = mock.Mock()
    mock_client_credentials_tokens.return_value = client_token_data
    mock_confidential_client.oauth2_client_credentials_tokens = (
        mock_client_credentials_tokens
    )

    factory = ClientCredentialsAuthorizerFactory(
        token_storage=mock_token_storage,
        confidential_client=mock_confidential_client,
        scope_requirements={"rs1": ["scope1"]},
    )

    # calling get_authorizer once creates a new authorizer automatically making
    # a client credentials call to get an access token that is then stored
    authorizer = factory.get_authorizer("rs1")
    assert authorizer.get_authorization_header() == "Bearer rs1_access_token_1"
    assert mock_client_credentials_tokens.call_count == 1


def test_client_credentials_authorizer_factory_no_scopes():
    mock_token_storage = _make_mem_token_storage()
    mock_confidential_client = mock.Mock()
    factory = ClientCredentialsAuthorizerFactory(
        token_storage=mock_token_storage,
        confidential_client=mock_confidential_client,
        scope_requirements={},
    )

    with pytest.raises(ValueError) as exc:
        factory.get_authorizer("rs2")
    assert (
        str(exc.value)
        == "ValidatingTokenStorage has no scope_requirements for resource_server rs2"
    )



from __future__ import annotations

import typing as t
import uuid

from globus_sdk._missing import MISSING, MissingType
from globus_sdk._payload import GlobusPayload


class GCSRoleDocument(GlobusPayload):
    """
    Convenience class for constructing a Role document
    to use as the `data` parameter to `create_role`

    :param DATA_TYPE: Versioned document type.
    :param collection: Collection ID for the collection the role will apply to.
        This value is omitted when creating an endpoint
        role or when creating role definitions when creating collections.
    :param principal: Auth identity or group id URN. Should be in the format
        urn:globus:auth:[identity|group]:{uuid of identity or group}
    :param role: Role assigned to the principal. Known values are owner,
        administrator, access_manager, activity_manager, and activity_monitor
    """

    def __init__(
        self,
        DATA_TYPE: str = "role#1.0.0",
        collection: uuid.UUID | str | MissingType = MISSING,
        principal: str | MissingType = MISSING,
        role: str | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = DATA_TYPE
        self["collection"] = collection
        self["principal"] = principal
        self["role"] = role
        self.update(additional_fields or {})



from __future__ import annotations

import dataclasses
import typing as t

import globus_sdk
from globus_sdk.config import get_environment_name
from globus_sdk.login_flows import (
    CommandLineLoginFlowManager,
    LocalServerLoginFlowManager,
    LoginFlowManager,
)
from globus_sdk.token_storage import (
    JSONTokenStorage,
    MemoryTokenStorage,
    SQLiteTokenStorage,
    TokenStorage,
    TokenValidationError,
)
from globus_sdk.token_storage.validating_token_storage import IdentityMismatchError

from .protocols import (
    IDTokenDecoderProvider,
    LoginFlowManagerProvider,
    TokenStorageProvider,
    TokenValidationErrorHandler,
)

if t.TYPE_CHECKING:
    from .app import GlobusApp


KnownLoginFlowManager = t.Literal["command-line", "local-server"]
KNOWN_LOGIN_FLOW_MANAGERS: dict[KnownLoginFlowManager, LoginFlowManagerProvider] = {
    "command-line": CommandLineLoginFlowManager,
    "local-server": LocalServerLoginFlowManager,
}
KnownTokenStorage = t.Literal["json", "sqlite", "memory"]
KNOWN_TOKEN_STORAGES: dict[KnownTokenStorage, t.Type[TokenStorageProvider]] = {
    "json": JSONTokenStorage,
    "sqlite": SQLiteTokenStorage,
    "memory": MemoryTokenStorage,
}


def resolve_by_login_flow(app: GlobusApp, error: TokenValidationError) -> None:
    """
    An error handler for GlobusApp token access errors that will retry the
    login flow if the error is a TokenValidationError.

    :param app: The GlobusApp instance which encountered an error.
    :param error: The encountered token validation error.
    """
    if isinstance(error, IdentityMismatchError):
        # An identity mismatch error indicates incorrect use of the app. Not something
        #   that can be resolved by running a login flow.
        raise error

    app.login(force=True)


@dataclasses.dataclass(frozen=True)
class GlobusAppConfig:
    """
    An immutable dataclass used to control the behavior of a :class:`GlobusApp`.

    :ivar bool request_refresh_tokens: Whether to request ``refresh tokens`` (expire
        after 6 months of no use) or use exclusively ``access tokens`` (expire 2 hours
        after issuance). Default: ``False``.
    :ivar bool auto_redrive_gares: If true, Globus Authorization Required Errors (GAREs)
        encountered during service interaction will automatically trigger a login flow
        to obtain new tokens and retry the failed request. Default: ``False``.

    :ivar str | ``TokenStorage`` | ``TokenStorageProvider`` token_storage:
        A class responsible for storing and retrieving tokens.
        This may be either a well-known provider (one of
        :class:`"json" <globus_sdk.token_storage.JSONTokenStorage>`,
        :class:`"sqlite" <globus_sdk.token_storage.SQLiteTokenStorage>`, or
        :class:`"memory" <globus_sdk.token_storage.MemoryTokenStorage>`) or a custom
        storage/provider. Default: ``"json"``.

    :ivar str | ``LoginFlowManager`` | ``LoginFlowManagerProvider`` login_flow_manager:
        A class responsible for overseeing Globus Auth login flows.
        This may be either be a well-known provider (one of
        :class:`"command-line" <globus_sdk.login_flows.CommandLineLoginFlowManager>` or
        :class:`"local-server" <globus_sdk.login_flows.LocalServerLoginFlowManager>`)
        or a custom manager/provider. Default: ``"command-line"``.

        .. note::

            **login_flow_manager** may be ignored when using a :class:`ClientApp`.

    :ivar str | None login_redirect_uri: The destination for Globus Auth to send
        a user after once completed a login flow. Default: ``None``.

        .. note::

            **login_redirect_url** may be ignored when using a
            :class:`NativeAppAuthClient`.
            Explicit values must be pre-registered on your client
            `here <https://app.globus.org/settings/developers>`_.

    :ivar ``TokenValidationErrorHandler`` | None token_validation_error_handler: A
        handler invoked to resolve errors raised during token validation. Set this to
        ``None`` to disable auto-login on service token validation errors.
        Default: ``resolve_by_login_flow`` (runs a login flow, storing the resulting
        tokens).

    :ivar ``IDTokenDecoder`` | ``IDTokenDecoderProvider`` id_token_decoder:
        An ID token decoder or a provider which produces a decoder. The decoder is used
        when decoding ``id_token`` JWTs from Globus Auth.
        Defaults to ``IDTokenDecoder``.

    :ivar str environment: The Globus environment of services to interact with. This is
        mostly used for testing purposes. This may additionally be set with the
        environment variable `GLOBUS_SDK_ENVIRONMENT`. Default: ``"production"``.
    """

    login_flow_manager: (
        KnownLoginFlowManager | LoginFlowManagerProvider | LoginFlowManager | None
    ) = None
    login_redirect_uri: str | None = None
    token_storage: KnownTokenStorage | TokenStorageProvider | TokenStorage = "json"
    request_refresh_tokens: bool = False
    auto_redrive_gares: bool = False
    token_validation_error_handler: TokenValidationErrorHandler | None = (
        resolve_by_login_flow
    )
    id_token_decoder: globus_sdk.IDTokenDecoder | IDTokenDecoderProvider = (
        globus_sdk.IDTokenDecoder
    )
    environment: str = dataclasses.field(default_factory=get_environment_name)


DEFAULT_CONFIG: GlobusAppConfig = GlobusAppConfig()



import globus_sdk
from globus_sdk.testing import load_response


def test_get_result_amqp_url(compute_client_v2: globus_sdk.ComputeClientV2):
    load_response(compute_client_v2.get_result_amqp_url)
    res = compute_client_v2.get_result_amqp_url()
    assert res.http_status == 200
    assert "connection_url" in res.data



from __future__ import annotations

import dataclasses
import sys
import typing as t
from collections import defaultdict, deque

from .errors import ScopeCycleError, ScopeParseError

SPECIAL_CHARACTERS = set("[]* ")
SPECIAL_TOKENS = set("[]*")


class ScopeGraph:
    def __init__(self) -> None:
        self.top_level_scopes: set[tuple[str, bool]] = set()
        self.nodes: set[str] = set()
        self.edges: set[tuple[str, str, bool]] = set()
        self.adjacency_matrix: dict[str, set[tuple[str, str, bool]]] = defaultdict(set)

    def breadth_first_walk(self) -> t.Iterator[tuple[str, bool]]:
        """
        Do a BFS across the forest, returning nodes (as tuples) in BFS order.

        For inspection of the graph after parsing.
        """
        bfs_queue: t.Deque[tuple[str, bool]] = deque(self.top_level_scopes)

        while bfs_queue:
            yield (current := bfs_queue.popleft())

            edges = self.adjacency_matrix[current[0]]
            for _, dest, optional in edges:
                bfs_queue.append((dest, optional))

    def add_edge(self, src: str, dest: str, optional: bool) -> None:
        self.edges.add((src, dest, optional))
        self.adjacency_matrix[src].add((src, dest, optional))

    def _normalize_optionals(self) -> None:
        to_remove: set[tuple[str, str, bool]] = set()
        for edge in self.edges:
            src, dest, optional = edge
            if not optional:
                continue
            # The current edge is optional; see if it's superseded by required edge
            required_variant = (src, dest, False)
            if required_variant in self.edges:
                to_remove.add(edge)
        self.edges = self.edges - to_remove
        for edge in to_remove:
            src, _, _ = edge
            self.adjacency_matrix[src].remove(edge)

    def _check_cycles(self) -> None:
        # explore the graph using an iterative Depth-First Search
        # as we explore the graph, keep track of paths of ancestry being explored
        # if we ever find a back-edge along one of those paths of ancestry, that
        # means that there must be a cycle

        # start from the top-level nodes (which we know to be the roots of this
        # forest-shaped graph)
        # we will track this as the set of paths to continue to branch and explore in a
        # stack and pop from it until it is empty, thus implementing DFS
        #
        # conceptually, the paths could be implemented as `list[str]`, which would
        # preserve the order in which we encountered each node. Using a set is a
        # micro-optimization which makes checks faster, since we only care to detect
        # *that* there was a cycle, not what the shape of that cycle was
        paths_to_explore: list[tuple[set[str], str]] = [
            ({node}, node) for node, _ in self.top_level_scopes
        ]

        while paths_to_explore:
            path, terminus = paths_to_explore.pop()

            # get out-edges from the last node in the path
            children = self.adjacency_matrix[terminus]

            # if the node was a leaf, no children, we are done exploring this path
            if not children:
                continue

            # for each child edge, do two basic things:
            # - check if we found a back-edge (cycle!)
            # - create a new path to explore, with the child node as its current
            #   terminus
            for edge in children:
                _, dest, _ = edge
                if dest in path:
                    raise ScopeCycleError(f"A cycle was found involving '{dest}'")
                paths_to_explore.append((path.union((dest,)), dest))

    def __str__(self) -> str:
        lines = ["digraph scopes {", '  rankdir="LR";', ""]
        for node, optional in self.top_level_scopes:
            lines.append(f"  {'*' if optional else ''}{node}")
        lines.append("")

        # do two passes to put all non-optional edges first
        for source, dest, optional in self.edges:
            if optional:
                continue
            lines.append(f"  {source} -> {dest};")
        for source, dest, optional in self.edges:
            if not optional:
                continue
            lines.append(f'  {source} -> {dest} [ label = "optional" ];')
        lines.append("")
        lines.append("}")
        return "\n".join(lines)

    @classmethod
    def parse(cls, scopes: str) -> ScopeGraph:
        trees = ScopeTreeNode.parse(scopes)
        graph = cls._convert_trees(trees)
        graph._normalize_optionals()
        graph._check_cycles()
        return graph

    @classmethod
    def _convert_trees(cls, trees: list[ScopeTreeNode]) -> ScopeGraph:
        graph = ScopeGraph()
        node_queue: t.Deque[ScopeTreeNode] = deque()

        for tree_node in trees:
            node_queue.append(tree_node)
            graph.top_level_scopes.add((tree_node.scope_string, tree_node.optional))

        while node_queue:
            tree_node = node_queue.pop()
            scope_string = tree_node.scope_string
            graph.nodes.add(scope_string)
            for dep in tree_node.dependencies:
                node_queue.append(dep)
                graph.add_edge(scope_string, dep.scope_string, dep.optional)

        return graph


# pass slots=True on 3.10+
# it's not strictly necessary, but it improves performance
if sys.version_info >= (3, 10):
    _add_dataclass_kwargs: dict[str, bool] = {"slots": True}
else:
    _add_dataclass_kwargs: dict[str, bool] = {}


@dataclasses.dataclass(**_add_dataclass_kwargs)
class ScopeTreeNode:
    #
    # This is an intermediate representation for scope parsing.
    #
    scope_string: str
    optional: bool
    dependencies: list[ScopeTreeNode] = dataclasses.field(default_factory=list)

    def add_dependency(self, subtree: ScopeTreeNode) -> None:
        self.dependencies.append(subtree)

    @staticmethod
    def parse(scope_string: str) -> list[ScopeTreeNode]:
        tokens = _tokenize(scope_string)
        return _parse_tokens(tokens)


def _tokenize(scope_string: str) -> list[str]:
    tokens: list[str] = []
    start = 0
    for idx, current_char, next_char in _peek_enumerate(scope_string):
        if current_char not in SPECIAL_CHARACTERS:
            continue
        _reject_bad_adjacent_characters(current_char, next_char)

        if start != idx:
            tokens.append(scope_string[start:idx])
        start = idx + 1

        if current_char in SPECIAL_TOKENS:
            tokens.append(current_char)
        elif current_char == " ":
            pass
        else:
            raise NotImplementedError
    remainder = scope_string[start:].strip()
    if remainder:
        tokens.append(remainder)
    return tokens


def _reject_bad_adjacent_characters(current_char: str, next_char: str | None) -> None:
    """Given a pair of adjacent characters during tokenization, raise
    appropriate errors if they are not valid."""
    if next_char is None:
        return

    if (current_char, next_char) == ("*", " "):
        raise ScopeParseError("'*' must not be followed by a space")
    elif current_char == "]" and next_char not in (" ", "]"):
        raise ScopeParseError("']' may only be followed by a space or ']'")
    elif (current_char, next_char) == (" ", "["):
        raise ScopeParseError("'[' cannot have a preceding space")


def _parse_tokens(tokens: list[str]) -> list[ScopeTreeNode]:
    # value to return
    ret: list[ScopeTreeNode] = []
    # track whether or not the current scope is optional (has a preceding *)
    current_optional = False
    # keep a stack of "parents", each time we enter a `[` context, push the last scope
    # and each time we exit via a `]`, pop from the stack
    parents: list[ScopeTreeNode] = []
    # track the current (or, by similar terminology, "last") complete scope seen
    current_scope: ScopeTreeNode | None = None

    for _, token, next_token in _peek_enumerate(tokens):
        _reject_bad_adjacent_tokens(token, next_token)

        if token == "*":
            current_optional = True
        elif token == "[":
            if not current_scope:
                raise ScopeParseError("found '[' without a preceding scope string")

            parents.append(current_scope)
        elif token == "]":
            if not parents:
                raise ScopeParseError("found ']' with no matching '[' preceding it")
            parents.pop()
        else:
            current_scope = ScopeTreeNode(token, optional=current_optional)
            current_optional = False
            if parents:
                parents[-1].add_dependency(current_scope)
            else:
                ret.append(current_scope)
    if parents:
        raise ScopeParseError("unclosed brackets, missing ']'")

    return ret


def _reject_bad_adjacent_tokens(current_token: str, next_token: str | None) -> None:
    """
    Given a pair of tokens from parsing, raise appropriate errors if they are
    not a valid sequence.
    """
    if current_token == "*":
        if next_token is None:
            raise ScopeParseError("ended in optional marker")
        elif next_token in SPECIAL_TOKENS:
            raise ScopeParseError(
                "a scope string must always follow an optional marker"
            )
    elif (current_token, next_token) == ("[", None):
        raise ScopeParseError("ended in left bracket")
    elif (current_token, next_token) == ("[", "]"):
        raise ScopeParseError("found empty brackets")
    elif (current_token, next_token) == ("[", "["):
        raise ScopeParseError("found double left-bracket")


def _peek_enumerate(data: str | list[str]) -> t.Iterator[tuple[int, str, str | None]]:
    """
    An iterator producing (index, character, next_char)
    or else producing (index, str, next_str)

    (Depending on whether or not the input is a string or list of strings)
    """
    if not data:
        return

    prev: str = data[0]
    for idx, c in enumerate(data[1:]):
        yield (idx, prev, c)
        prev = c

    yield (len(data) - 1, prev, None)



from __future__ import annotations

from types import SimpleNamespace
from uuid import UUID

import pytest

from globus_sdk.scopes.consents import ConsentForest, ConsentTreeConstructionError
from tests.common import ConsentTest, ScopeRepr

_zero_uuid = str(UUID(int=0))


def _uuid_of(char: str) -> str:
    if len(char) != 1:
        raise ValueError(f"char must be a single character, got {char!r}")
    return _zero_uuid.replace("0", char)


Clients = SimpleNamespace(
    Zero=_uuid_of("0"),
    One=_uuid_of("1"),
    Two=_uuid_of("2"),
    Three=_uuid_of("3"),
)
Scopes = SimpleNamespace(
    A=ScopeRepr(_uuid_of("A"), "A"),
    B=ScopeRepr(_uuid_of("B"), "B"),
    C=ScopeRepr(_uuid_of("C"), "C"),
    D=ScopeRepr(_uuid_of("D"), "D"),
)


def test_consent_forest_creation():
    root = ConsentTest.of(Clients.Zero, Scopes.A)
    node1 = ConsentTest.of(Clients.One, Scopes.B, parent=root)
    node2 = ConsentTest.of(Clients.Two, Scopes.C, parent=node1)

    forest = ConsentForest([root, node1, node2])
    assert len(forest.trees) == 1
    tree = forest.trees[0]
    assert tree.root == root
    assert tree.max_depth == 3

    assert tree.edges[tree.root.id] == {node1.id}
    assert tree.edges[node1.id] == {node2.id}
    assert tree.edges[node2.id] == set()

    assert tree.get_node(root.id) == root
    assert tree.get_node(node1.id) == node1
    assert tree.get_node(node2.id) == node2


def test_consent_forest_scope_requirement_evaluation():
    root = ConsentTest.of(Clients.Zero, Scopes.A)
    node1 = ConsentTest.of(Clients.One, Scopes.B, parent=root)
    node2 = ConsentTest.of(Clients.Two, Scopes.C, parent=node1)

    forest = ConsentForest([root, node1, node2])

    assert forest.meets_scope_requirements("A")
    assert forest.meets_scope_requirements("A[B[C]]")
    assert not forest.meets_scope_requirements("B")
    assert not forest.meets_scope_requirements("A[C]")


def test_consent_forest_scope_requirement_with_sibling_dependent_scopes():
    root = ConsentTest.of(Clients.Zero, Scopes.A)
    node1 = ConsentTest.of(Clients.One, Scopes.B, parent=root)
    node2 = ConsentTest.of(Clients.Two, Scopes.C, parent=root)

    forest = ConsentForest([root, node1, node2])

    assert forest.meets_scope_requirements("A")
    assert forest.meets_scope_requirements("A[B]")
    assert forest.meets_scope_requirements("A[C]")
    assert forest.meets_scope_requirements("A[B C]")
    assert not forest.meets_scope_requirements("A[B[C]]")
    assert not forest.meets_scope_requirements("A[C[B]]")


@pytest.mark.parametrize("atomically_revocable", (True, False))
def test_consent_forest_scope_requirement_with_optional_dependent_scopes(
    atomically_revocable: bool,
):
    """
    Dependent scope optionality is intentionally ignored for this implementation.

    In formal terms, the scope "A[*B]" is only satisfied by a tree matching the shape
      A -> B where B is "atomically revocable".
    We've decided that this is an auth service concern, not a concern for local clients
      to be making decisions about; so we intentionally ignore this distinction in order
      to give standard users a simpler verification mechanism to ask "will my request
      work with the current set of consents?".
    """
    root = ConsentTest.of(Clients.Zero, Scopes.A)
    child = ConsentTest.of(
        Clients.One, Scopes.B, parent=root, atomically_revocable=atomically_revocable
    )

    forest = ConsentForest([root, child])

    assert forest.meets_scope_requirements("A[B]")
    assert forest.meets_scope_requirements("A[*B]")


def test_consent_forest_with_disjoint_consents_with_duplicate_scopes():
    """
    Strange state to reproduce in practice but this test case simulates the forest of
      Tree 1: A (Client Zero) -> B (Client Zero)
      Tree 2: B (Client Zero) -> C (Client Zero)

    In this situation, A[B] and B[C] are both satisfied, but A[B[C]] is not.
    """
    root1 = ConsentTest.of(Clients.Zero, Scopes.A)
    child1 = ConsentTest.of(Clients.Zero, Scopes.B, parent=root1)

    root2 = ConsentTest.of(Clients.Zero, Scopes.B)
    child2 = ConsentTest.of(Clients.Zero, Scopes.C, parent=root2)

    forest = ConsentForest([root1, child1, root2, child2])

    assert forest.meets_scope_requirements("A[B]")
    assert forest.meets_scope_requirements("B[C]")
    assert not forest.meets_scope_requirements("A[B[C]]")


def test_consent_forest_with_missing_intermediary_nodes():
    """
    Simulate a situation in which we didn't receive the full list of consents from
      Auth. So the tree has holes

    Tree: A -> <B should be here but isn't> -> C
    """
    root = ConsentTest.of(Clients.Zero, Scopes.A)
    node1 = ConsentTest.of(Clients.One, Scopes.B, parent=root)
    node2 = ConsentTest.of(Clients.Two, Scopes.C, parent=node1)

    # Only add the first and last node to the forest.
    # The last node (C) references the middle node (B) and so forest loading should
    #   fail.
    with pytest.raises(
        ConsentTreeConstructionError, match=rf"Missing parent node: {node1.id}"
    ):
        ConsentForest([root, node2])



from tests.common import register_api_route_fixture_file


def test_update_group(groups_client):
    register_api_route_fixture_file(
        "groups",
        "/v2/groups/592e0566-5201-4207-b5e1-7cd6c516e9a0",
        "updated_group.json",
        method="PUT",
    )

    data = {
        "name": "Claptrap's Rough Riders",
        "description": "Stairs strongly discouraged.",
    }
    res = groups_client.update_group(
        group_id="592e0566-5201-4207-b5e1-7cd6c516e9a0", data=data
    )
    assert res.http_status == 200
    assert "Claptrap" in res.data["name"]
    assert "Stairs strongly discouraged." in res.data["description"]



from copy import deepcopy

from responses.matchers import query_param_matcher

from globus_sdk.testing import RegisteredResponse, ResponseList, ResponseSet

from ._common import FLOW_DESCRIPTION, RUN, RUN_ID

RUN_WITH_FLOW_DESCRIPTION = deepcopy(RUN)
RUN_WITH_FLOW_DESCRIPTION["flow_description"] = FLOW_DESCRIPTION

RESPONSES = ResponseSet(
    metadata={"run_id": RUN_ID},
    default=ResponseList(
        RegisteredResponse(
            service="flows",
            method="GET",
            path=f"/runs/{RUN_ID}",
            json=RUN,
            match=[query_param_matcher(params={})],
        ),
        RegisteredResponse(
            service="flows",
            method="GET",
            path=f"/runs/{RUN_ID}",
            json=RUN,
            match=[query_param_matcher(params={"include_flow_description": "False"})],
        ),
        RegisteredResponse(
            service="flows",
            method="GET",
            path=f"/runs/{RUN_ID}",
            json=RUN_WITH_FLOW_DESCRIPTION,
            match=[query_param_matcher(params={"include_flow_description": "True"})],
        ),
    ),
)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

CREDENTIAL_ID = "af43d884-64a1-4414-897a-680c32374439"

RESPONSES = ResponseSet(
    metadata={"id": CREDENTIAL_ID},
    default=RegisteredResponse(
        service="gcs",
        path="/user_credentials",
        method="POST",
        json={
            "DATA_TYPE": "result#1.0.0",
            "code": "success",
            "data": [
                {
                    "DATA_TYPE": "user_credential#1.0.0",
                    "connector_id": "145812c8-decc-41f1-83cf-bb2a85a2a70b",
                    "display_name": "updated_posix_credential",
                    "id": CREDENTIAL_ID,
                    "identity_id": "948847d4-ffcc-4ae0-ba3a-a4c88d480159",
                    "invalid": False,
                    "policies": {"DATA_TYPE": "posix_user_credential_policies#1.0.0"},
                    "provisioned": False,
                    "storage_gateway_id": "82247cc9-3208-4d71-bd7f-1b8798c95e6b",
                    "username": "testuser",
                },
            ],
            "detail": "created",
            "has_next_page": False,
            "http_response_code": 201,
            "message": f"Created User Credential {CREDENTIAL_ID}",
        },
    ),
)



import pytest

from globus_sdk import (
    ActiveScaleStoragePolicies,
    AzureBlobStoragePolicies,
    BlackPearlStoragePolicies,
    BoxStoragePolicies,
    CephStoragePolicies,
    GoogleCloudStoragePolicies,
    GoogleDriveStoragePolicies,
    HPSSStoragePolicies,
    IrodsStoragePolicies,
    OneDriveStoragePolicies,
    POSIXStagingStoragePolicies,
    POSIXStoragePolicies,
    S3StoragePolicies,
    StorageGatewayDocument,
)
from globus_sdk.transport import JSONRequestEncoder


@pytest.mark.parametrize(
    "use_kwargs,doc_version",
    [
        ({}, "1.0.0"),
        ({"require_mfa": True}, "1.1.0"),
        ({"require_mfa": False}, "1.1.0"),
    ],
)
def test_datatype_version_deduction(use_kwargs, doc_version):
    sg = StorageGatewayDocument(**use_kwargs)
    assert sg["DATA_TYPE"] == f"storage_gateway#{doc_version}"


def test_storage_gateway_policy_document_conversion():
    policies = POSIXStoragePolicies(
        groups_allow=["jedi", "wookies"], groups_deny=["sith", "stormtroopers"]
    )
    sg = StorageGatewayDocument(policies=policies)
    assert "policies" in sg
    assert isinstance(sg["policies"], POSIXStoragePolicies)

    encoder = JSONRequestEncoder()
    request_data = encoder.encode("POST", "bogus.url.example", {}, sg, {}).json

    assert request_data["policies"] == {
        "DATA_TYPE": "posix_storage_policies#1.0.0",
        "groups_allow": ["jedi", "wookies"],
        "groups_deny": ["sith", "stormtroopers"],
    }


def test_posix_staging_env_vars():
    p = POSIXStagingStoragePolicies(
        groups_allow=("vulcans", "starfleet"),
        groups_deny=(x for x in ("ferengi", "romulans")),
        stage_app="/globus/bin/posix-stage-data",
        environment=({"name": "VOLUME", "value": "/vol/0"},),
    )

    assert isinstance(p["environment"], list)
    assert dict(p) == {
        "DATA_TYPE": "posix_staging_storage_policies#1.0.0",
        "groups_allow": ["vulcans", "starfleet"],
        "groups_deny": ["ferengi", "romulans"],
        "stage_app": "/globus/bin/posix-stage-data",
        "environment": [{"name": "VOLUME", "value": "/vol/0"}],
    }


@pytest.mark.parametrize(
    "doc_class",
    [
        StorageGatewayDocument,
        POSIXStagingStoragePolicies,
        POSIXStoragePolicies,
        BlackPearlStoragePolicies,
        BoxStoragePolicies,
        CephStoragePolicies,
        GoogleDriveStoragePolicies,
        GoogleCloudStoragePolicies,
        OneDriveStoragePolicies,
        AzureBlobStoragePolicies,
        S3StoragePolicies,
        ActiveScaleStoragePolicies,
        IrodsStoragePolicies,
        HPSSStoragePolicies,
    ],
)
def test_storage_gateway_documents_support_additional_fields(doc_class):
    d = doc_class()
    assert "DATA_TYPE" in d
    assert "foo" not in d

    d2 = doc_class(additional_fields={"foo": "bar"})
    assert "DATA_TYPE" in d2
    assert d2["foo"] == "bar"



from functools import cached_property

from ..collection import DynamicScopeCollection, _url_scope, _urn_scope
from ..representation import Scope


class GCSEndpointScopes(DynamicScopeCollection):
    """
    A dynamic ScopeCollection with a named property for the GCS
    manage_collections scope. "manage_collections" is a scope on GCS Endpoints.
    The resource_server string should be the GCS Endpoint ID.

    **Examples**

    >>> sc = GCSEndpointScopes("xyz")
    >>> mc_scope = sb.manage_collections
    """

    _scope_names = ("manage_collections",)

    @cached_property
    def manage_collections(self) -> Scope:
        return _urn_scope(self.resource_server, "manage_collections")


class GCSCollectionScopes(DynamicScopeCollection):
    """
    A dynamic ScopeCollection with a named property for the GCS data_access scope.
    "data_access" is a scope on GCS Collections. The resource_server string should
    be the GCS Collection ID.

    **Examples**

    >>> sc = GCSCollectionScopes("xyz")
    >>> da_scope = sc.data_access
    >>> https_scope = sc.https
    """

    _scope_names = ("data_access", "https")

    @cached_property
    def data_access(self) -> Scope:
        return _url_scope(self.resource_server, "data_access")

    @cached_property
    def https(self) -> Scope:
        return _url_scope(self.resource_server, "https")



import json
import urllib.parse
import uuid

import pytest
import responses

import globus_sdk
from globus_sdk._missing import filter_missing
from globus_sdk.testing import get_last_request, load_response
from tests.common import register_api_route_fixture_file


@pytest.fixture
def search_client():
    client = globus_sdk.SearchClient()
    with client.retry_config.tune(max_retries=0):
        yield client


def test_search_query_simple(search_client):
    meta = load_response(search_client.search).metadata

    res = search_client.search(meta["index_id"], q="foo")
    assert res.http_status == 200

    data = res.data
    assert isinstance(data, dict)
    assert data["gmeta"][0]["entries"][0]["content"]["foo"] == "bar"

    req = get_last_request()
    assert req.body is None
    parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(req.url).query)
    assert parsed_qs == {"q": ["foo"]}


@pytest.mark.parametrize("query_doc", [{"q": "foo"}, {"q": "foo", "limit": 10}])
def test_search_post_query_simple(search_client, query_doc):
    meta = load_response(search_client.post_search).metadata

    res = search_client.post_search(meta["index_id"], query_doc)
    assert res.http_status == 200

    data = res.data
    assert isinstance(data, dict)
    assert data["gmeta"][0]["entries"][0]["content"]["foo"] == "bar"

    req = get_last_request()
    assert req.body is not None
    req_body = json.loads(req.body)
    assert req_body == dict(query_doc)


def test_search_post_query_simple_with_v1_helper(search_client):
    query_doc = globus_sdk.SearchQueryV1(q="foo")
    meta = load_response(search_client.post_search).metadata

    res = search_client.post_search(meta["index_id"], query_doc)
    assert res.http_status == 200

    data = res.data
    assert isinstance(data, dict)
    assert data["gmeta"][0]["entries"][0]["content"]["foo"] == "bar"

    req = get_last_request()
    assert req.body is not None
    req_body = json.loads(req.body)
    assert req_body == {"@version": "query#1.0.0", "q": "foo"}


@pytest.mark.parametrize("doc_type", ("dict", "helper"))
def test_search_post_query_arg_overrides(search_client, doc_type):
    meta = load_response(search_client.post_search).metadata

    if doc_type == "dict":
        query_doc = {"q": "foo", "limit": 10, "offset": 0}
    elif doc_type == "helper":
        query_doc = globus_sdk.SearchQueryV1(q="foo", limit=10, offset=0)
    else:
        raise NotImplementedError(doc_type)
    res = search_client.post_search(meta["index_id"], query_doc, limit=100, offset=150)
    assert res.http_status == 200

    data = res.data
    assert isinstance(data, dict)
    assert data["gmeta"][0]["entries"][0]["content"]["foo"] == "bar"

    req = get_last_request()
    assert req.body is not None
    req_body = json.loads(req.body)
    assert req_body != dict(query_doc)
    assert req_body["q"] == query_doc["q"]
    assert req_body["limit"] == 100
    assert req_body["offset"] == 150
    # important! these should be unchanged (no side-effects)
    assert query_doc["limit"] == 10
    assert query_doc["offset"] == 0


@pytest.mark.parametrize(
    "query_doc",
    [
        {"q": "foo"},
        globus_sdk.SearchScrollQuery("foo"),
    ],
)
def test_search_paginated_scroll_query(search_client, query_doc):
    index_id = str(uuid.uuid1())
    register_api_route_fixture_file(
        "search",
        f"/v1/index/{index_id}/scroll",
        "scroll_result_1.json",
        method="POST",
        match=[responses.matchers.json_params_matcher({"q": "foo"})],
    )
    register_api_route_fixture_file(
        "search",
        f"/v1/index/{index_id}/scroll",
        "scroll_result_2.json",
        method="POST",
        match=[
            responses.matchers.json_params_matcher(
                {"q": "foo", "marker": "3d34900e3e4211ebb0a806b2af333354"}
            )
        ],
    )

    data = list(search_client.paginated.scroll(index_id, query_doc).items())
    assert len(responses.calls) == 2
    assert len(data) == 2

    assert isinstance(data[0], dict)
    assert data[0]["entries"][0]["content"]["foo"] == "bar"

    assert isinstance(data[1], dict)
    assert data[1]["entries"][0]["content"]["foo"] == "baz"

    # confirm that pagination was not side-effecting
    assert "marker" not in filter_missing(query_doc)



from .client import FlowsClient, SpecificFlowClient
from .data import RunActivityNotificationPolicy
from .errors import FlowsAPIError
from .response import (
    IterableFlowsResponse,
    IterableRegisteredAPIsResponse,
    IterableRunLogsResponse,
    IterableRunsResponse,
)

__all__ = (
    "FlowsAPIError",
    "FlowsClient",
    "IterableFlowsResponse",
    "IterableRegisteredAPIsResponse",
    "IterableRunLogsResponse",
    "IterableRunsResponse",
    "SpecificFlowClient",
    "RunActivityNotificationPolicy",
)



from __future__ import annotations

import datetime
import typing as t

# these types are aliases meant for internal use
IntLike = t.Union[int, str]
DateLike = t.Union[str, datetime.datetime]


class ResponseLike(t.Protocol):
    @property
    def http_status(self) -> int: ...

    @property
    def http_reason(self) -> str: ...

    @property
    def headers(self) -> t.Mapping[str, str]: ...

    @property
    def content_type(self) -> str | None: ...

    @property
    def text(self) -> str: ...

    @property
    def binary_content(self) -> bytes: ...


class Closable(t.Protocol):
    def close(self) -> None: ...



import globus_sdk
import globus_sdk.gare

# this is the SDK tutorial client ID, replace with your own ID
NATIVE_CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"
# set the collection ID to your test collection
COLLECTION_ID = "..."


def print_ls_data(client):
    ls_result = client.operation_ls(COLLECTION_ID)
    for item in ls_result:
        name = item["name"]
        if item["type"] == "dir":
            name += "/"
        print(name)


with globus_sdk.UserApp("ls-session", client_id=NATIVE_CLIENT_ID) as app:
    client = globus_sdk.TransferClient(app=app)

    # because the recommended test configuration uses a mapped collection
    # without High Assurance capabilities, it will have a data_access scope
    # requirement
    # comment out this line if your collection does not use data_access
    client.add_app_data_access_scope(COLLECTION_ID)

    # try to run the desired operation (`print_ls_data`)
    try:
        print_ls_data(client)
    # catch the possible API error
    except globus_sdk.TransferAPIError as err:
        # if there are authorization parameters data in the error,
        # use it to redrive login
        if err.info.authorization_parameters:
            print("An authorization requirement was not met. Logging in again...")

            gare = globus_sdk.gare.to_gare(err)
            params = gare.authorization_parameters
            # set 'prompt=login', which guarantees a fresh login without
            # reliance on the browser session
            params.prompt = "login"

            # pass these parameters into a login flow
            app.login(auth_params=params)

            # rerun the desired print_ls_data() operation
            print_ls_data(client)

        # otherwise, there are no authorization parameters, so reraise the error
        else:
            raise



from __future__ import annotations

import typing as t

import globus_sdk

from .base import StorageAdapter


class MemoryAdapter(StorageAdapter):
    """
    A token storage adapter which stores tokens in process memory.

    Tokens are lost when the process exits.
    """

    def __init__(self) -> None:
        self._tokens: dict[str, dict[str, t.Any]] = {}

    def store(self, token_response: globus_sdk.OAuthTokenResponse) -> None:
        self._tokens.update(token_response.by_resource_server)

    def get_token_data(self, resource_server: str) -> dict[str, t.Any] | None:
        return self._tokens.get(resource_server)



from .client import TransferClientV2
from .data import (
    BookmarkCreateDocument,
    BookmarkUpdateDocument,
    TunnelCreateDocument,
    TunnelUpdateDocument,
)

__all__ = (
    "BookmarkCreateDocument",
    "BookmarkUpdateDocument",
    "TransferClientV2",
    "TunnelCreateDocument",
    "TunnelUpdateDocument",
)



import urllib.parse

from globus_sdk.response import ArrayResponse
from globus_sdk.testing import get_last_request, load_response


def test_get_my_groups(groups_client):
    meta = load_response(groups_client.get_my_groups).metadata

    res = groups_client.get_my_groups(statuses="active")
    assert res.http_status == 200

    req = get_last_request()
    parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(req.url).query)
    assert parsed_qs["statuses"] == ["active"]

    assert isinstance(res, ArrayResponse)
    assert isinstance(res.data, list)
    assert set(meta["group_names"]) == {g["name"] for g in res}



import uuid


def _as_uuid(s: str) -> str:
    return str(uuid.UUID(int=int(s, 36)))


SUBMISSION_ID = _as_uuid("submission_id")
ENDPOINT_ID = _as_uuid("endpoint_id")
TASK_ID = _as_uuid("task_id")
SUBSCRIPTION_ID = _as_uuid("subscription_id")



import copy

from globus_sdk.testing import RegisteredResponse, ResponseSet

from ._common import TWO_HOP_TRANSFER_RUN

RUN_ID = TWO_HOP_TRANSFER_RUN["run_id"]

SUCCESSFUL_DELETE_RESPONSE = copy.deepcopy(TWO_HOP_TRANSFER_RUN)
SUCCESSFUL_DELETE_RESPONSE.update(
    {
        "status": "SUCCEEDED",
        "display_status": "SUCCEEDED",
        "details": {
            "code": "FlowSucceeded",
            "output": {},
            "description": "The Flow run reached a successful completion state",
        },
    }
)

CONFLICT_RESPONSE = {
    "error": {
        "code": "STATE_CONFLICT",
        "detail": (
            f"Run {RUN_ID} has status 'ACTIVE' but must have status"
            f" in {{'ENDED', 'SUCCEEDED', 'FAILED'}} for requested operation"
        ),
    },
    "debug_id": "80d920b6-66cf-4254-bcbd-7d3efe814e1a",
}

RESPONSES = ResponseSet(
    metadata={"run_id": RUN_ID},
    default=RegisteredResponse(
        service="flows",
        method="POST",
        path=f"/runs/{RUN_ID}/release",
        json=SUCCESSFUL_DELETE_RESPONSE,
    ),
    conflict=RegisteredResponse(
        service="flows",
        method="POST",
        path=f"/runs/{RUN_ID}/release",
        status=409,
        json=CONFLICT_RESPONSE,
    ),
)



"""
The GCSDownloader provides HTTPS file download capabilities for Globus Connect Server.
"""

from __future__ import annotations

import logging
import sys
import types
import typing as t
import urllib.parse

import globus_sdk
import globus_sdk.transport
from globus_sdk._internal.type_definitions import Closable
from globus_sdk.experimental.gcs_collection_client import GCSCollectionClient
from globus_sdk.transport.default_retry_checks import DEFAULT_RETRY_CHECKS

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

log = logging.getLogger(__name__)


class GCSDownloader:
    """
    An object which manages connection and authentication state to enable HTTPS
    downloads from a specific Globus Connect Server collection.

    The initial request to read a file features support for determining authentication
    requirements dynamically, and subsequent requests will reuse that authentication
    data.

    Using a single :class:`GCSDownloader` to access distinct collections is not
    supported. A separate downloader should be used for each collection.

    Downloaders may be used as context managers, in which case they automatically call
    their ``close()`` method on exit:

    >>> with GCSDownloader(app) as downloader:
    >>>     print(downloader.read_file(url))

    :param app: The :class:`GlobusApp` used to authenticate calls to this server.
    :param gcs_client: The underlying client used for the file read request. Typically
        omitted. When not provided, one will be constructed on demand by the downloader.
    :param transfer_client: A client used when detecting collection information.
        Typically omitted. When not provided, one will be constructed on demand by the
        downloader.
    :param transport: A transport for the downloader, used for authentication
        sniffing operations. When a client is built by the downloader it will
        inherit this transport.
    """

    def __init__(
        self,
        app: globus_sdk.GlobusApp,
        *,
        gcs_client: GCSCollectionClient | None = None,
        transfer_client: globus_sdk.TransferClient | None = None,
        transport: globus_sdk.transport.RequestsTransport | None = None,
    ) -> None:
        self.app = app
        self._resources_to_close: list[Closable] = []

        if transport is not None:
            self.transport = transport
        else:
            self.transport = globus_sdk.transport.RequestsTransport()
            self._resources_to_close.append(self.transport)

        # the downloader will need a RetryConfig when it uses its own transport
        self._retry_config = globus_sdk.transport.RetryConfig()
        self._retry_config.checks.register_many_checks(DEFAULT_RETRY_CHECKS)

        self.gcs_client = gcs_client

        # set the transfer_client if provided
        self.transfer_client = transfer_client

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> None:
        self.close()

    def close(self) -> None:
        """
        Close all resources which are owned by this downloader.
        """
        for resource in self._resources_to_close:
            log.debug(
                f"closing resource of type {type(resource).__name__} "
                f"for {type(self).__name__}"
            )
            resource.close()

    @t.overload
    def read_file(self, file_uri: str, *, as_text: t.Literal[True]) -> str: ...
    @t.overload
    def read_file(self, file_uri: str, *, as_text: t.Literal[False]) -> bytes: ...
    @t.overload
    def read_file(self, file_uri: str) -> str: ...

    def read_file(self, file_uri: str, *, as_text: bool = True) -> str | bytes:
        """
        Given a file URI on a GCS Collection, read the data.

        :param file_uri: The full URI of the file on the collection which is being
            downloaded.
        :param as_text: When ``True``, the file contents are decoded into a string. Set
            to ``False`` to retrieve data as bytes.

        .. caution::

            The file read is done naively as a GET request. This may be unsuitable for
            very large files.
        """
        # dynamically build a client if needed
        if self.gcs_client is None:
            self.gcs_client = self._get_client_from_uri(file_uri)
            self._resources_to_close.append(self.gcs_client)

        response = self.gcs_client.get(file_uri)
        if as_text:
            return response.text
        return response.binary_content

    def _get_client_from_uri(self, file_uri: str) -> GCSCollectionClient:
        collection_id = self._sniff_collection_id(file_uri)
        collection_address = _get_collection_address(file_uri)

        client = self._gcs_client_constructor(
            collection_client_id=collection_id,
            collection_address=collection_address,
        )
        self.app.add_scope_requirements(
            {client.scopes.resource_server: self._determine_required_scopes(client)}
        )
        return client

    def _gcs_client_constructor(
        self, *, collection_client_id: str, collection_address: str
    ) -> GCSCollectionClient:
        return GCSCollectionClient(
            collection_client_id,
            collection_address,
            app=self.app,
            transport=self.transport,
        )

    def _determine_required_scopes(
        self, client: GCSCollectionClient
    ) -> list[globus_sdk.Scope]:
        if self.transfer_client is None:
            self.transfer_client = globus_sdk.TransferClient(
                app=self.app, transport=self.transport
            )
            self._resources_to_close.append(self.transfer_client)

        required_scopes: list[globus_sdk.Scope] = [client.scopes.https]
        if _uses_data_access(self.transfer_client, client.collection_id):
            required_scopes.append(client.scopes.data_access)
        return required_scopes

    def _sniff_collection_id(self, file_uri: str) -> str:
        response = self.transport.request(
            "GET",
            file_uri,
            caller_info=globus_sdk.transport.RequestCallerInfo(
                retry_config=self._retry_config
            ),
            allow_redirects=False,
        )
        if "Location" not in response.headers:
            msg = (
                f"Attempting to detect the collection ID for the file at '{file_uri}' "
                "failed. Did not receive a redirect with Location header on "
                "unauthenticated call."
            )
            raise RuntimeError(msg)

        location_header = response.headers["Location"]
        parsed_location = urllib.parse.urlparse(location_header)
        parsed_location_qs = urllib.parse.parse_qs(parsed_location.query)

        if "client_id" not in parsed_location_qs:
            msg = (
                f"Attempting to detect the collection ID for the file at '{file_uri}' "
                "failed. Location header did not encode a 'client_id'."
            )
            raise RuntimeError(msg)

        client_ids = parsed_location_qs["client_id"]
        if len(client_ids) != 1:
            msg = (
                f"Attempting to detect the collection ID for the file at '{file_uri}' "
                "failed. Multiple 'client_id' params were present."
            )
            raise RuntimeError(msg)

        return client_ids[0]


def _get_collection_address(file_uri: str) -> str:
    parsed = urllib.parse.urlparse(file_uri)
    return f"{parsed.scheme}://{parsed.netloc}"


def _uses_data_access(
    transfer_client: globus_sdk.TransferClient, collection_id: str
) -> bool:
    doc = transfer_client.get_endpoint(collection_id)
    if doc["entity_type"] != "GCSv5_mapped_collection":
        return False
    if doc["high_assurance"]:
        return False
    return True



from __future__ import annotations

import logging
import typing as t
import uuid

from globus_sdk._missing import MISSING, MissingType
from globus_sdk._payload import GlobusPayload

log = logging.getLogger(__name__)


class CreateTunnelData(GlobusPayload):
    """
    Convenience class for constructing a tunnel document to use as the
    ``data`` parameter to
    :meth:`create_tunnel <globus_sdk.TransferClient.create_tunnel>`.
    """

    def __init__(
        self,
        initiator_stream_access_point: uuid.UUID | str,
        listener_stream_access_point: uuid.UUID | str,
        *,
        label: str | MissingType = MISSING,
        submission_id: uuid.UUID | str | MissingType = MISSING,
        lifetime_mins: int | MissingType = MISSING,
        restartable: bool | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        log.debug("Creating a new CreateTunnelData object")

        relationships = {
            "listener": {
                "data": {
                    "type": "StreamAccessPoint",
                    "id": listener_stream_access_point,
                }
            },
            "initiator": {
                "data": {
                    "type": "StreamAccessPoint",
                    "id": initiator_stream_access_point,
                }
            },
        }
        attributes = {
            "label": label,
            "submission_id": submission_id,
            "restartable": restartable,
            "lifetime_mins": lifetime_mins,
        }
        if additional_fields is not None:
            attributes.update(additional_fields)

        self["data"] = {
            "type": "Tunnel",
            "relationships": relationships,
            "attributes": attributes,
        }



from globus_sdk.testing import get_last_request, load_response


def test_get_tunnel_events(client):
    meta = load_response(client.get_tunnel_events).metadata
    res = client.get_tunnel_events(meta["tunnel_id"])
    assert res.http_status == 200
    assert res["data"][0]["type"] == "TunnelEvent"
    assert res["data"][1]["type"] == "TunnelEvent"

    req = get_last_request()
    assert req.body is None



#!/usr/bin/env python
import argparse
import os
import sys

import globus_sdk
from globus_sdk.token_storage import SimpleJSONFileAdapter

MY_FILE_ADAPTER = SimpleJSONFileAdapter(os.path.expanduser("~/.sdk-manage-flow.json"))

# tutorial client ID
# we recommend replacing this with your own client for any production use-cases
CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"

NATIVE_CLIENT = globus_sdk.NativeAppAuthClient(CLIENT_ID)


def do_login_flow(scope):
    NATIVE_CLIENT.oauth2_start_flow(requested_scopes=scope, refresh_tokens=True)
    authorize_url = NATIVE_CLIENT.oauth2_get_authorize_url()
    print(f"Please go to this URL and login:\n\n{authorize_url}\n")
    auth_code = input("Please enter the code here: ").strip()
    tokens = NATIVE_CLIENT.oauth2_exchange_code_for_tokens(auth_code)
    return tokens


def get_authorizer(flow_id=None):
    if flow_id:
        resource_server = flow_id
        scope = globus_sdk.SpecificFlowClient(flow_id).scopes.user
    else:
        resource_server = globus_sdk.FlowsClient.resource_server
        scope = globus_sdk.FlowsClient.scopes.manage_flows

    # try to load the tokens from the file, possibly returning None
    if MY_FILE_ADAPTER.file_exists():
        tokens = MY_FILE_ADAPTER.get_token_data(resource_server)
    else:
        tokens = None

    if tokens is None:
        # do a login flow, getting back initial tokens
        response = do_login_flow(scope)
        # now store the tokens and pull out the correct token
        MY_FILE_ADAPTER.store(response)
        tokens = response.by_resource_server[resource_server]

    return globus_sdk.RefreshTokenAuthorizer(
        tokens["refresh_token"],
        NATIVE_CLIENT,
        access_token=tokens["access_token"],
        expires_at=tokens["expires_at_seconds"],
        on_refresh=MY_FILE_ADAPTER.on_refresh,
    )


def get_flows_client():
    return globus_sdk.FlowsClient(authorizer=get_authorizer())


def get_specific_flow_client(flow_id):
    authorizer = get_authorizer(flow_id)
    return globus_sdk.SpecificFlowClient(flow_id, authorizer=authorizer)


def create_flow(args):
    flows_client = get_flows_client()
    print(
        flows_client.create_flow(
            title=args.title,
            definition={
                "StartAt": "DoIt",
                "States": {
                    "DoIt": {
                        "Type": "Action",
                        "ActionUrl": "https://actions.globus.org/hello_world",
                        "Parameters": {
                            "echo_string": "Hello, Asynchronous World!",
                        },
                        "End": True,
                    }
                },
            },
            input_schema={},
            subtitle="A flow created by the SDK tutorial",
        )
    )


def delete_flow(args):
    flows_client = get_flows_client()
    print(flows_client.delete_flow(args.flow_id))


def list_flows():
    flows_client = get_flows_client()
    for flow in flows_client.list_flows(filter_roles="flow_owner"):
        print(f"title: {flow['title']}")
        print(f"id: {flow['id']}")
        print()


def run_flow(args):
    flow_client = get_specific_flow_client(args.flow_id)
    print(flow_client.run_flow({}))


def logout():
    for tokendata in MY_FILE_ADAPTER.get_by_resource_server().values():
        for tok_key in ("access_token", "refresh_token"):
            token = tokendata[tok_key]
            NATIVE_CLIENT.oauth2_revoke_token(token)

    os.remove(MY_FILE_ADAPTER.filename)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("action", choices=["logout", "create", "delete", "list", "run"])
    parser.add_argument("-f", "--flow-id", help="Flow ID for delete and run")
    parser.add_argument("-t", "--title", help="Name for create")
    args = parser.parse_args()

    try:
        if args.action == "logout":
            logout()
        elif args.action == "create":
            if args.title is None:
                parser.error("create requires --title")
            create_flow(args)
        elif args.action == "delete":
            if args.flow_id is None:
                parser.error("delete requires --flow-id")
            delete_flow(args)
        elif args.action == "list":
            list_flows()
        elif args.action == "run":
            if args.flow_id is None:
                parser.error("run requires --flow-id")
            run_flow(args)
        else:
            raise NotImplementedError()
    except globus_sdk.FlowsAPIError as e:
        print(f"API Error: {e.code} {e.message}")
        print(e.text)
        sys.exit(1)


if __name__ == "__main__":
    main()



from __future__ import annotations

import typing as t

from globus_sdk._missing import MISSING, MissingType
from globus_sdk._payload import GlobusPayload


class SearchQueryV1(GlobusPayload):
    """
    A specialized dict which has helpers for creating and modifying a Search
    Query document.

    :param q: The query string. Required unless filters are used.
    :param limit: A limit on the number of results returned in a single page
    :param offset: An offset into the set of all results for the query
    :param advanced: Whether to enable (``True``) or not to enable (``False``) advanced
        parsing of query strings. The default of ``False`` is robust and guarantees that
        the query will not error with "bad query string" errors
    :param filters: a list of filters to apply to the query
    :param facets: a list of facets to apply to the query
    :param post_facet_filters: a list of filters to apply after facet
        results are returned
    :param boosts: a list of boosts to apply to the query
    :param sort: a list of fields to sort results
    :param additional_fields: additional data to include in the query document
    """

    def __init__(
        self,
        *,
        q: str | MissingType = MISSING,
        limit: int | MissingType = MISSING,
        offset: int | MissingType = MISSING,
        advanced: bool | MissingType = MISSING,
        filters: list[dict[str, t.Any]] | MissingType = MISSING,
        facets: list[dict[str, t.Any]] | MissingType = MISSING,
        post_facet_filters: list[dict[str, t.Any]] | MissingType = MISSING,
        boosts: list[dict[str, t.Any]] | MissingType = MISSING,
        sort: list[dict[str, t.Any]] | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self["@version"] = "query#1.0.0"

        self["q"] = q
        self["limit"] = limit
        self["offset"] = offset
        self["advanced"] = advanced
        self["filters"] = filters
        self["facets"] = facets
        self["post_facet_filters"] = post_facet_filters
        self["boosts"] = boosts
        self["sort"] = sort
        self.update(additional_fields or {})


class SearchScrollQuery(GlobusPayload):
    """
    A scrolling query type, for scrolling the full result set for an index.

    Scroll queries have more limited capabilities than general searches.
    They cannot boost fields, sort, or apply facets. They can, however, still apply the
    same filtering mechanisms which are available to normal queries.

    Scrolling also differs in that it supports the use of the ``marker`` field, which is
    used to paginate results.

    :param q: The query string
    :param limit: A limit on the number of results returned in a single page
    :param advanced: Whether to enable (``True``) or not to enable (``False``) advanced
        parsing of query strings. The default of ``False`` is robust and guarantees that
        the query will not error with "bad query string" errors
    :param marker: the marker value
    :param additional_fields: additional data to include in the query document
    """

    def __init__(
        self,
        q: str | MissingType = MISSING,
        *,
        limit: int | MissingType = MISSING,
        advanced: bool | MissingType = MISSING,
        marker: str | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self["q"] = q
        self["limit"] = limit
        self["advanced"] = advanced
        self["marker"] = marker
        self.update(additional_fields or {})



from __future__ import annotations

import json
import logging
import typing as t

import jwt
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey

from globus_sdk._missing import MISSING, MissingType
from globus_sdk.response import GlobusHTTPResponse

log = logging.getLogger(__name__)


class _JWKGetCallbackProto(t.Protocol):
    def __call__(
        self,
        path: str,
        *,
        query_params: dict[str, t.Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> GlobusHTTPResponse: ...


def get_jwk_data(
    *,
    fget: _JWKGetCallbackProto,
    openid_configuration: GlobusHTTPResponse | dict[str, t.Any],
) -> dict[str, t.Any]:
    jwks_uri = openid_configuration["jwks_uri"]
    log.debug("fetching from jwks_uri=%s", jwks_uri)
    data = fget(jwks_uri).data
    if not isinstance(data, dict):
        # how could this happen?
        # some guesses:
        # - interfering proxy or cache
        # - user passed explicit (incorrect) OIDC config
        raise ValueError(
            "JWK data was not a dict. This should be an unreachable condition."
        )
    return data


def pem_decode_jwk_data(
    *,
    jwk_data: dict[str, t.Any],
) -> RSAPublicKey:
    log.debug("JWK PEM decode requested, decoding...")
    # decode from JWK to an RSA PEM key for JWT decoding
    # cast here because this should never be private key
    jwk_as_pem: RSAPublicKey = t.cast(
        RSAPublicKey,
        jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(jwk_data["keys"][0])),
    )
    log.debug("JWK PEM decoding finished successfully")
    return jwk_as_pem


@t.runtime_checkable
class SupportsJWKMethods(t.Protocol):
    client_id: str | None

    def get_openid_configuration(self) -> GlobusHTTPResponse: ...

    @t.overload
    def get_jwk(
        self,
        openid_configuration: GlobusHTTPResponse | dict[str, t.Any] | MissingType,
        *,
        as_pem: t.Literal[True],
    ) -> RSAPublicKey: ...

    @t.overload
    def get_jwk(
        self,
        openid_configuration: GlobusHTTPResponse | dict[str, t.Any] | MissingType,
        *,
        as_pem: t.Literal[False],
    ) -> dict[str, t.Any]: ...

    def get_jwk(
        self,
        openid_configuration: (
            GlobusHTTPResponse | dict[str, t.Any] | MissingType
        ) = MISSING,
        *,
        as_pem: bool = False,
    ) -> RSAPublicKey | dict[str, t.Any]: ...



from globus_sdk.token_storage import MemoryTokenStorage


def test_store_and_get_token_data_by_resource_server(
    mock_token_data_by_resource_server,
):
    adapter = MemoryTokenStorage()
    adapter.store_token_data_by_resource_server(mock_token_data_by_resource_server)

    gotten = adapter.get_token_data_by_resource_server()

    for resource_server in ["resource_server_1", "resource_server_2"]:
        assert (
            mock_token_data_by_resource_server[resource_server].to_dict()
            == gotten[resource_server].to_dict()
        )


def test_store_token_response_with_namespace(mock_response):
    adapter = MemoryTokenStorage(namespace="foo")
    adapter.store_token_response(mock_response)

    assert (
        adapter._tokens["foo"]["resource_server_1"]["access_token"] == "access_token_1"
    )
    assert (
        adapter._tokens["foo"]["resource_server_2"]["access_token"] == "access_token_2"
    )


def test_get_token_data(mock_response):
    adapter = MemoryTokenStorage()
    adapter.store_token_response(mock_response)

    assert adapter.get_token_data("resource_server_1").access_token == "access_token_1"
    assert adapter.get_token_data("resource_server_2").access_token == "access_token_2"


def test_remove_token_data(mock_response):
    adapter = MemoryTokenStorage()
    adapter.store_token_response(mock_response)

    # remove rs1, confirm only rs2 is still available
    remove_result = adapter.remove_token_data("resource_server_1")
    assert remove_result is True

    assert adapter.get_token_data("resource_server_1") is None
    assert adapter.get_token_data("resource_server_2").access_token == "access_token_2"

    # confirm unable to re-remove rs1
    remove_result = adapter.remove_token_data("resource_server_1")
    assert remove_result is False



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import TIMER_ID, V1_TIMER

UPDATED_NAME = "updated name"
UPDATED_TIMER = dict(V1_TIMER)
UPDATED_TIMER["name"] = UPDATED_NAME  # mypy complains if this is onelinerized

RESPONSES = ResponseSet(
    metadata={"job_id": TIMER_ID, "name": UPDATED_NAME},
    default=RegisteredResponse(
        service="timer",
        path=f"/jobs/{TIMER_ID}",
        method="PATCH",
        json=UPDATED_TIMER,
    ),
)



import time
from unittest import mock

import pytest

from globus_sdk import exc
from globus_sdk.authorizers.renewing import EXPIRES_ADJUST_SECONDS, RenewingAuthorizer


class MockRenewer(RenewingAuthorizer):
    """
    Class that implements RenewingAuthorizer so that _get_token_response and
    _extract_token_data can return known values for testing
    """

    def __init__(self, token_data, **kwargs) -> None:
        self.token_data = token_data
        self.token_response = mock.Mock()
        super().__init__(**kwargs)

    def _get_token_response(self):
        return self.token_response

    def _extract_token_data(self, res):
        return self.token_data


ACCESS_TOKEN = "access_token_1"


@pytest.fixture
def expires_at():
    return int(time.time()) + EXPIRES_ADJUST_SECONDS + 10


@pytest.fixture
def token_data():
    return {
        "expires_at_seconds": int(time.time()) + 1000,
        "access_token": "access_token_2",
    }


@pytest.fixture
def on_refresh():
    return mock.Mock()


@pytest.fixture
def authorizer(on_refresh, token_data, expires_at):
    return MockRenewer(
        token_data,
        access_token=ACCESS_TOKEN,
        expires_at=expires_at,
        on_refresh=on_refresh,
    )


@pytest.fixture
def expired_authorizer(on_refresh, token_data, expires_at):
    return MockRenewer(
        token_data,
        access_token=ACCESS_TOKEN,
        expires_at=expires_at - 11,
        on_refresh=on_refresh,
    )


def test_init(token_data, expires_at):
    """
    Creating a MockRenewer with partial data (expires_at, access_token) results in
    errors.
    Either complete data or no partial data works.
    """
    authorizer = MockRenewer(
        token_data, access_token=ACCESS_TOKEN, expires_at=expires_at
    )
    assert authorizer.access_token == ACCESS_TOKEN
    assert authorizer.access_token != token_data["access_token"]

    # with no args, an automatic "refresh" is triggered on init
    authorizer = MockRenewer(token_data)
    assert authorizer.access_token == token_data["access_token"]
    assert authorizer.expires_at == token_data["expires_at_seconds"]

    with pytest.raises(exc.GlobusSDKUsageError):
        MockRenewer(token_data, access_token=ACCESS_TOKEN)

    with pytest.raises(exc.GlobusSDKUsageError):
        MockRenewer(token_data, expires_at=expires_at)


def test_get_new_access_token(authorizer, token_data, on_refresh):
    """
    Calls get_new_access token, confirms that the mock _get_token_data
    is used and that the mock on_refresh function is called.
    """
    # take note of original access_token_hash
    original_hash = authorizer._access_token_hash

    # get new_access_token
    authorizer._get_new_access_token()
    # confirm side effects
    assert authorizer.access_token == token_data["access_token"]
    assert authorizer.expires_at == token_data["expires_at_seconds"]
    assert authorizer._access_token_hash != original_hash
    on_refresh.assert_called_once()


def test_ensure_valid_token_ok(authorizer):
    """
    Confirms nothing is done before the access_token expires,
    """
    authorizer.ensure_valid_token()
    assert authorizer.access_token == ACCESS_TOKEN


def test_ensure_valid_token_expired(expired_authorizer, token_data):
    """
    Confirms a new access_token is gotten after expiration
    """
    expired_authorizer.ensure_valid_token()
    assert expired_authorizer.access_token == token_data["access_token"]
    assert expired_authorizer.expires_at == token_data["expires_at_seconds"]


def test_ensure_valid_token_no_token(authorizer, token_data):
    """
    Confirms a new access_token is gotten if the old one is set to None
    """
    authorizer.access_token = None
    authorizer.ensure_valid_token()
    assert authorizer.access_token == token_data["access_token"]
    assert authorizer.expires_at == token_data["expires_at_seconds"]


def test_ensure_valid_token_no_expiration(authorizer, token_data):
    """
    Confirms a new access_token is gotten if expires_at is set to None
    """
    authorizer.expires_at = None
    authorizer.ensure_valid_token()
    assert authorizer.access_token == token_data["access_token"]
    assert authorizer.expires_at == token_data["expires_at_seconds"]


def test_get_authorization_header(authorizer):
    """
    Gets authorization header, confirms expected value
    """
    assert authorizer.get_authorization_header() == "Bearer " + ACCESS_TOKEN


def test_get_authorization_header_expired(expired_authorizer, token_data):
    """
    Sets the access_token to be expired, then gets authorization header
    Confirms header value uses the new access_token.
    """
    assert expired_authorizer.get_authorization_header() == (
        "Bearer " + token_data["access_token"]
    )


def test_get_authorization_header_no_token(authorizer, token_data):
    """
    Sets the access_token to None, then gets authorization header
    Confirms header value uses the new access_token.
    """
    authorizer.access_token = None
    assert authorizer.get_authorization_header() == (
        "Bearer " + token_data["access_token"]
    )


def test_get_authorization_header_no_expires(authorizer, token_data):
    """
    Sets expires_at to None, then gets authorization header
    Confirms header value uses the new access_token.
    """
    authorizer.expires_at = None
    assert authorizer.get_authorization_header() == (
        "Bearer " + token_data["access_token"]
    )


def test_handle_missing_authorization(authorizer):
    """
    Confirms that RenewingAuthorizers will attempt to fix 401s
    by treating their existing access_token as expired
    """
    assert authorizer.handle_missing_authorization()
    assert authorizer.expires_at is None



from .client import SearchClient
from .data import SearchQueryV1, SearchScrollQuery
from .errors import SearchAPIError

__all__ = (
    "SearchClient",
    "SearchQueryV1",
    "SearchScrollQuery",
    "SearchAPIError",
)



#!/usr/bin/env python

import argparse
import os

import globus_sdk
from globus_sdk.token_storage import SimpleJSONFileAdapter

MY_FILE_ADAPTER = SimpleJSONFileAdapter(
    os.path.expanduser("~/.sdk-manage-projects.json")
)

SCOPES = [globus_sdk.AuthClient.scopes.manage_projects, "openid", "email"]
RESOURCE_SERVER = globus_sdk.AuthClient.resource_server

# tutorial client ID
# we recommend replacing this with your own client for any production use-cases
CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"

NATIVE_CLIENT = globus_sdk.NativeAppAuthClient(CLIENT_ID)


def do_login_flow():
    NATIVE_CLIENT.oauth2_start_flow(requested_scopes=SCOPES)
    authorize_url = NATIVE_CLIENT.oauth2_get_authorize_url()
    print(f"Please go to this URL and login:\n\n{authorize_url}\n")
    auth_code = input("Please enter the code here: ").strip()
    tokens = NATIVE_CLIENT.oauth2_exchange_code_for_tokens(auth_code)
    return tokens.by_resource_server[RESOURCE_SERVER]


def get_auth_client():
    tokens = do_login_flow()
    return globus_sdk.AuthClient(
        authorizer=globus_sdk.AccessTokenAuthorizer(tokens["access_token"])
    )


def create_project(args):
    auth_client = get_auth_client()
    userinfo = auth_client.userinfo()
    print(
        auth_client.create_project(
            args.name,
            contact_email=userinfo["email"],
            admin_ids=userinfo["sub"],
        )
    )


def list_projects():
    auth_client = get_auth_client()
    for project in auth_client.get_projects():
        print(f"name: {project['display_name']}")
        print(f"id: {project['id']}")
        print()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("action", choices=["create", "list"])
    parser.add_argument("-n", "--name", help="Project name for create")
    args = parser.parse_args()

    execute(parser, args)


def execute(parser, args):
    if args.action == "create":
        if args.name is None:
            parser.error("create requires --name")
        create_project(args)
    elif args.action == "list":
        list_projects()
    else:
        raise NotImplementedError()


if __name__ == "__main__":
    main()



from ..collection import StaticScopeCollection, _urn_scope


class _SearchScopes(StaticScopeCollection):
    resource_server = "search.api.globus.org"

    all = _urn_scope(resource_server, "all")
    globus_connect_server = _urn_scope(resource_server, "globus_connect_server")
    ingest = _urn_scope(resource_server, "ingest")
    search = _urn_scope(resource_server, "search")


SearchScopes = _SearchScopes()



"""
Tests for submitting Transfer and Delete tasks
"""

import json

import pytest

from globus_sdk import DeleteData, TransferAPIError, TransferData
from globus_sdk.testing import get_last_request, load_response
from tests.common import GO_EP1_ID, GO_EP2_ID


def test_transfer_submit_failure(client):
    load_response(client.get_submission_id)
    meta = load_response(client.submit_transfer, case="failure").metadata

    with pytest.raises(TransferAPIError) as excinfo:
        client.submit_transfer(TransferData(GO_EP1_ID, GO_EP2_ID))

    assert excinfo.value.http_status == 400
    assert excinfo.value.request_id == meta["request_id"]
    assert excinfo.value.code == "ClientError.BadRequest.NoTransferItems"


def test_transfer_submit_success(client):
    load_response(client.get_submission_id)
    meta = load_response(client.submit_transfer).metadata

    tdata = TransferData(
        GO_EP1_ID,
        GO_EP2_ID,
        label="mytask",
        sync_level="exists",
        deadline="2018-06-01",
        source_local_user="my-source-user",
        destination_local_user="my-dest-user",
        additional_fields={"custom_param": "foo"},
    )
    assert tdata["custom_param"] == "foo"
    assert tdata["sync_level"] == 0

    tdata.add_item("/path/to/foo", "/path/to/bar")

    res = client.submit_transfer(tdata)

    assert res
    assert res["submission_id"] == meta["submission_id"]
    assert res["task_id"] == meta["task_id"]

    req_body = json.loads(get_last_request().body)
    assert req_body["source_local_user"] == "my-source-user"
    assert req_body["destination_local_user"] == "my-dest-user"


def test_delete_submit_success(client):
    load_response(client.get_submission_id)
    meta = load_response(client.submit_delete).metadata

    ddata = DeleteData(
        endpoint=GO_EP1_ID,
        label="mytask",
        deadline="2018-06-01",
        local_user="my-user",
        additional_fields={"custom_param": "foo"},
    )
    assert ddata["custom_param"] == "foo"

    ddata.add_item("/path/to/foo")

    res = client.submit_delete(ddata)

    assert res
    assert res["submission_id"] == meta["submission_id"]
    assert res["task_id"] == meta["task_id"]

    req_body = json.loads(get_last_request().body)
    assert req_body["local_user"] == "my-user"


@pytest.mark.parametrize("datatype", ("transfer", "delete"))
def test_submit_adds_missing_submission_id_to_data(client, datatype):
    data = {}
    meta = load_response(client.get_submission_id).metadata
    if datatype == "transfer":
        load_response(client.submit_transfer)
        client.submit_transfer(data)
    else:
        load_response(client.submit_delete)
        client.submit_delete(data)
    assert "submission_id" in data
    assert data["submission_id"] == meta["submission_id"]
    req_body = json.loads(get_last_request().body)
    assert req_body == data


@pytest.mark.parametrize("datatype", ("transfer", "delete"))
def test_submit_does_not_overwrite_existing_submission_id(client, datatype):
    data = {"submission_id": "foo"}
    meta = load_response(client.get_submission_id).metadata
    if datatype == "transfer":
        load_response(client.submit_transfer)
        client.submit_transfer(data)
    else:
        load_response(client.submit_delete)
        client.submit_delete(data)
    assert data["submission_id"] == "foo"
    assert data["submission_id"] != meta["submission_id"]
    req_body = json.loads(get_last_request().body)
    assert req_body == data



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from .._common import ENDPOINT_ID

RESPONSES = ResponseSet(
    metadata={"endpoint_id": ENDPOINT_ID},
    default=RegisteredResponse(
        service="compute",
        path=f"/v2/endpoints/{ENDPOINT_ID}",
        method="DELETE",
        json={"result": 302},
    ),
)



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

ACCESS_POINT_ID_1 = str(uuid.uuid4())
ACCESS_POINT_ID_2 = str(uuid.uuid4())


RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="transfer",
        method="GET",
        path="/v2/stream_access_points",
        json={
            "data": [
                {
                    "attributes": {
                        "advertised_owner": "john@globus.org",
                        "contact_email": None,
                        "contact_info": None,
                        "department": None,
                        "description": None,
                        "display_name": "Buzz Dev Listener 1",
                        "info_link": None,
                        "keywords": None,
                        "organization": None,
                        "tlsftp_server": (
                            "tlsftp://s-463c7.e7d5e.8540."
                            "test3.zones.dnsteam.globuscs.info:443"
                        ),
                    },
                    "id": ACCESS_POINT_ID_1,
                    "relationships": {
                        "host_endpoint": {
                            "data": {
                                "id": "d6428474-c308-4a2d-8a86-d377915d978b",
                                "type": "Endpoint",
                            }
                        }
                    },
                    "type": "StreamAccessPoint",
                },
                {
                    "attributes": {
                        "advertised_owner": "john@globus.org",
                        "contact_email": None,
                        "contact_info": None,
                        "department": None,
                        "description": None,
                        "display_name": "Buzz Dev Listener 2",
                        "info_link": None,
                        "keywords": None,
                        "organization": None,
                        "tlsftp_server": (
                            "tlsftp://s-a845b.29a01.671e."
                            "test3.zones.dnsteam.globuscs.info:443"
                        ),
                    },
                    "id": ACCESS_POINT_ID_2,
                    "relationships": {
                        "host_endpoint": {
                            "data": {
                                "id": "d6428474-c308-4a2d-8a86-d377915d978b",
                                "type": "Endpoint",
                            }
                        }
                    },
                    "type": "StreamAccessPoint",
                },
            ],
            "meta": {"request_id": "55QRq2iBa"},
        },
        metadata={"access_point_ids": [ACCESS_POINT_ID_1, ACCESS_POINT_ID_2]},
    ),
)



from __future__ import annotations

import globus_sdk

# type is (str | None)
s: str | None = globus_sdk.BaseClient.resource_server
i: int = globus_sdk.BaseClient.resource_server  # type: ignore [assignment]

# holds on an instance as well
c = globus_sdk.BaseClient()
s = c.resource_server
i = c.resource_server  # type: ignore [assignment]

# check that data:list warns, but other types are okay
r = c.request("POST", "/foo", data="bar")
r = c.request("POST", "/foo", data={})
r = c.request("POST", "/foo", data=[])  # type: ignore [arg-type]



import uuid

import pytest

from globus_sdk import exc
from globus_sdk._internal import guards
from globus_sdk._internal.serializable import Serializable


@pytest.mark.parametrize(
    "value, typ, ok",
    [
        # passing
        ([], str, True),
        ([1, 2], int, True),
        (["1", ""], str, True),
        ([], list, True),
        ([[], [1, 2], ["foo"]], list, True),
        # failing
        ([1], str, False),
        (["foo"], int, False),
        ((1, 2), int, False),
        (list, list, False),
        (list, str, False),
        (["foo", 1], str, False),
        ([1, 2], list, False),
    ],
)
def test_list_of_guard(value, typ, ok):
    assert guards.is_list_of(value, typ) == ok


@pytest.mark.parametrize(
    "value, typ, ok",
    [
        # passing
        (None, str, True),
        ("foo", str, True),
        # failing
        (b"foo", str, False),
        ("", int, False),
        (type(None), str, False),
    ],
)
def test_opt_guard(value, typ, ok):
    assert guards.is_optional(value, typ) == ok


@pytest.mark.parametrize(
    "value, typ, ok",
    [
        # passing
        ([], str, True),
        ([], int, True),
        ([1, 2], int, True),
        (["1", ""], str, True),
        (None, str, True),
        # failing
        # NB: the guard checks `list[str] | None`, not `list[str | None]`
        ([None], str, False),
        (b"foo", str, False),
        ("", str, False),
        (type(None), str, False),
    ],
)
def test_opt_list_guard(value, typ, ok):
    assert guards.is_optional_list_of(value, typ) == ok


@pytest.mark.parametrize("value", (uuid.UUID(int=0), str(uuid.UUID(int=1))))
def test_uuidlike_ok(value):
    assert guards.validators.uuidlike("foo", value) == value


@pytest.mark.parametrize("value", (str(uuid.UUID(int=0))[:-1], ""))
def test_uuidlike_fails_value(value):
    with pytest.raises(
        exc.ValidationError, match="'foo' must be a valid UUID"
    ) as excinfo:
        guards.validators.uuidlike("foo", value)

    err = excinfo.value
    assert f"value='{value}'" in str(err)


@pytest.mark.parametrize("value", (object(), None, ["bar"]))
def test_uuidlike_fails_type(value):
    with pytest.raises(
        exc.ValidationError, match="'foo' must be a UUID or str"
    ) as excinfo:
        guards.validators.uuidlike("foo", value)

    err = excinfo.value
    assert f"value='{value}'" in str(err)


@pytest.mark.parametrize(
    "validator, value",
    (
        pytest.param(guards.validators.str_, "bar", id="str"),
        pytest.param(guards.validators.int_, 0, id="int-0"),
        pytest.param(guards.validators.int_, 1, id="int-1"),
        pytest.param(guards.validators.opt_str, "bar", id="opt_str-str"),
        pytest.param(guards.validators.opt_str, None, id="opt_str-None"),
        pytest.param(guards.validators.opt_bool, True, id="opt_bool-True"),
        pytest.param(guards.validators.opt_bool, False, id="opt_bool-False"),
        pytest.param(guards.validators.opt_bool, None, id="opt_bool-None"),
        pytest.param(guards.validators.str_list, [], id="str_list-empty"),
        pytest.param(guards.validators.str_list, ["foo"], id="str_list-onestr"),
        pytest.param(guards.validators.opt_str_list, [], id="opt_str_list-empty"),
        pytest.param(guards.validators.opt_str_list, ["foo"], id="opt_str_list-onestr"),
        pytest.param(guards.validators.opt_str_list, None, id="opt_str_list-None"),
        pytest.param(
            guards.validators.opt_str_list_or_commasep,
            [],
            id="opt_str_list_or_commasep-emptylist",
        ),
        pytest.param(
            guards.validators.opt_str_list_or_commasep,
            ["foo"],
            id="opt_str_list_or_commasep-list",
        ),
        pytest.param(
            guards.validators.opt_str_list_or_commasep,
            None,
            id="opt_str_list_or_commasep-None",
        ),
    ),
)
def test_simple_validator_passing(validator, value):
    assert validator("foo", value) == value


@pytest.mark.parametrize(
    "validator, value, match_message",
    (
        pytest.param(guards.validators.str_, 1, "'foo' must be a string", id="str-int"),
        pytest.param(
            guards.validators.str_, False, "'foo' must be a string", id="str-bool"
        ),
        pytest.param(
            guards.validators.str_, None, "'foo' must be a string", id="str-None"
        ),
        pytest.param(
            guards.validators.int_, "bar", "'foo' must be an int", id="int-str"
        ),
        pytest.param(
            guards.validators.opt_str,
            0,
            "'foo' must be a string or null",
            id="opt_str-int",
        ),
        pytest.param(
            guards.validators.opt_bool,
            0,
            "'foo' must be a bool or null",
            id="opt_bool-int",
        ),
        pytest.param(
            guards.validators.str_list,
            "x",
            "'foo' must be a list of strings",
            id="str_list-str",
        ),
        pytest.param(
            guards.validators.opt_str_list,
            "x",
            "'foo' must be a list of strings or null",
            id="opt_str_list-str",
        ),
        pytest.param(
            guards.validators.opt_str_list_or_commasep,
            0,
            "'foo' must be a list of strings or a comma-delimited string or null",
            id="opt_str_list_or_commasep-int",
        ),
    ),
)
def test_simple_validator_failing(validator, value, match_message):
    with pytest.raises(exc.ValidationError, match=match_message):
        validator("foo", value)


def test_instance_or_dict_validator_failing():
    class MyObj(Serializable):
        def __init__(self, *, extra=None) -> None:
            pass

    with pytest.raises(
        exc.ValidationError, match="'foo' must be a 'MyObj' object or a dictionary"
    ):
        guards.validators.instance_or_dict("foo", object(), MyObj)


def test_instance_or_dict_validator_pass_on_simple_instance():
    class MyObj(Serializable):
        def __init__(self, *, extra=None) -> None:
            pass

    x = MyObj()
    y = guards.validators.instance_or_dict("foo", x, MyObj)
    assert x is y


def test_instance_or_dict_validator_pass_on_simple_dict():
    class MyObj(Serializable):
        def __init__(self, *, extra=None) -> None:
            pass

    x = guards.validators.instance_or_dict("foo", {}, MyObj)
    assert isinstance(x, MyObj)


def test_strlist_or_commasep_splits_str():
    x = guards.validators.opt_str_list_or_commasep("foo", "foo,bar,baz")
    assert x == ["foo", "bar", "baz"]



import pytest

import globus_sdk
from globus_sdk.testing import load_response


def test_reopen_index(client):
    meta = load_response(client.reopen_index).metadata

    res = client.reopen_index(meta["index_id"])
    assert res.http_status == 200
    assert res["acknowledged"] is True


def test_reopen_index_already_open(client):
    meta = load_response(client.reopen_index, case="already_open").metadata

    with pytest.raises(globus_sdk.SearchAPIError) as excinfo:
        client.reopen_index(meta["index_id"])

    err = excinfo.value

    assert err.http_status == 409
    assert err.code == "Conflict.IncompatibleIndexStatus"



from globus_sdk.testing import load_response


def test_get_consents(service_client):
    meta = load_response(service_client.get_consents).metadata
    res = service_client.get_consents(meta["identity_id"])

    forest = res.to_forest()
    assert len(forest.nodes) == 2
    assert len(forest.trees) == 1
    tree = forest.trees[0]
    assert forest.trees[0].max_depth == 2
    assert tree.root.scope_name.endswith("transfer.api.globus.org:all")
    children = [tree.get_node(child_id) for child_id in tree.edges[tree.root.id]]
    assert len(children) == 1
    assert children[0].scope_name.endswith("data_access")



import json

from globus_sdk import ConnectorTable, UserCredentialDocument
from globus_sdk.testing import get_last_request, load_response


def test_get_user_credential_list(client):
    metadata = load_response(client.get_user_credential_list).metadata
    res = client.get_user_credential_list()

    assert len(list(res)) == 2
    # sanity check some fields
    assert res["DATA_TYPE"] == "result#1.0.0"
    for item in res:
        assert item["DATA_TYPE"] == "user_credential#1.0.0"
        assert item["id"] in metadata["ids"]
        assert "identity_id" in item
        assert "username" in item


def test_get_user_credential(client):
    metadata = load_response(client.get_user_credential).metadata
    uc_id = metadata["id"]
    res = client.get_user_credential(uc_id)

    assert res["DATA_TYPE"] == "user_credential#1.0.0"
    assert res.full_data["DATA_TYPE"] == "result#1.0.0"
    assert res["id"] == uc_id
    assert res["display_name"] == "posix_credential"

    connector = ConnectorTable.lookup(res["connector_id"])
    assert connector is not None
    assert connector.name == "POSIX"


def test_create_user_credential(client):
    metadata = load_response(client.create_user_credential).metadata
    uc_id = metadata["id"]

    data = UserCredentialDocument(
        storage_gateway_id="82247cc9-3208-4d71-bd7f-1b8798c95e6b",
        identity_id="948847d4-ffcc-4ae0-ba3a-a4c88d480159",
        username="testuser",
        display_name="s3_credential",
        policies={
            "DATA_TYPE": "s3_user_credential_policies#1.0.0",
            "s3_key_id": "foo",
            "s3_secret_key": "bar",
        },
    )

    res = client.create_user_credential(data)
    assert res["DATA_TYPE"] == "user_credential#1.0.0"
    assert res.full_data["DATA_TYPE"] == "result#1.0.0"
    assert res.full_data["message"] == f"Created User Credential {uc_id}"

    req_body = req_body = json.loads(get_last_request().body)
    assert req_body["DATA_TYPE"] == "user_credential#1.0.0"
    assert req_body["policies"]["DATA_TYPE"] == "s3_user_credential_policies#1.0.0"
    for key, value in req_body.items():
        assert data[key] == value


def test_update_user_credential(client):
    metadata = load_response(client.update_user_credential).metadata
    uc_id = metadata["id"]

    data = UserCredentialDocument(
        display_name="updated_posix_credential",
    )

    res = client.update_user_credential(uc_id, data)
    assert res["DATA_TYPE"] == "user_credential#1.0.0"
    assert res.full_data["DATA_TYPE"] == "result#1.0.0"
    assert res.full_data["message"] == f"Updated User Credential {uc_id}"

    req_body = json.loads(get_last_request().body)
    assert req_body["display_name"] == "updated_posix_credential"


def test_delete_user_credential(client):
    metadata = load_response(client.delete_user_credential).metadata
    uc_id = metadata["id"]

    res = client.delete_user_credential(uc_id)
    assert res["DATA_TYPE"] == "result#1.0.0"
    assert res["message"] == f"Deleted User Credential {uc_id}"



import argparse

import globus_sdk
from globus_sdk.scopes import TransferScopes

parser = argparse.ArgumentParser()
parser.add_argument("SRC")
parser.add_argument("DST")
args = parser.parse_args()

CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"
auth_client = globus_sdk.NativeAppAuthClient(CLIENT_ID)


# we will need to do the login flow potentially twice, so define it as a
# function
#
# we default to using the Transfer "all" scope, but it is settable here
# look at the ConsentRequired handler below for how this is used
def login_and_get_transfer_client(*, scopes=TransferScopes.all):
    # note that 'requested_scopes' can be a single scope or a list
    # this did not matter in previous examples but will be leveraged in
    # this one
    auth_client.oauth2_start_flow(requested_scopes=scopes)
    authorize_url = auth_client.oauth2_get_authorize_url()
    print(f"Please go to this URL and login:\n\n{authorize_url}\n")

    auth_code = input("Please enter the code here: ").strip()
    tokens = auth_client.oauth2_exchange_code_for_tokens(auth_code)
    transfer_tokens = tokens.by_resource_server["transfer.api.globus.org"]

    # return the TransferClient object, as the result of doing a login
    return globus_sdk.TransferClient(
        authorizer=globus_sdk.AccessTokenAuthorizer(transfer_tokens["access_token"])
    )


# get an initial client to try with, which requires a login flow
transfer_client = login_and_get_transfer_client()

# now, try an ls on the source and destination to see if ConsentRequired
# errors are raised
consent_required_scopes = []


def check_for_consent_required(target):
    try:
        transfer_client.operation_ls(target, path="/")
    # catch all errors and discard those other than ConsentRequired
    # e.g. ignore PermissionDenied errors as not relevant
    except globus_sdk.TransferAPIError as err:
        if err.info.consent_required:
            consent_required_scopes.extend(err.info.consent_required.required_scopes)


check_for_consent_required(args.SRC)
check_for_consent_required(args.DST)

# the block above may or may not populate this list
# but if it does, handle ConsentRequired with a new login
if consent_required_scopes:
    print(
        "One of your endpoints requires consent in order to be used.\n"
        "You must login a second time to grant consents.\n\n"
    )
    transfer_client = login_and_get_transfer_client(scopes=consent_required_scopes)

# from this point onwards, the example is exactly the same as the reactive
# case, including the behavior to retry on ConsentRequiredErrors. This is
# not obvious, but there are cases in which it is necessary -- for example,
# if a user consents at the start, but the process of building task_data is
# slow, they could revoke their consent before the submission step
#
# in the common case, a single submission with no retry would suffice

task_data = globus_sdk.TransferData(
    source_endpoint=args.SRC, destination_endpoint=args.DST
)
task_data.add_item(
    "/share/godata/file1.txt",  # source
    "/~/example-transfer-script-destination.txt",  # dest
)


def do_submit(client):
    task_doc = client.submit_transfer(task_data)
    task_id = task_doc["task_id"]
    print(f"submitted transfer, task_id={task_id}")


try:
    do_submit(transfer_client)
except globus_sdk.TransferAPIError as err:
    if not err.info.consent_required:
        raise
    print(
        "Encountered a ConsentRequired error.\n"
        "You must login a second time to grant consents.\n\n"
    )
    transfer_client = login_and_get_transfer_client(
        scopes=err.info.consent_required.required_scopes
    )
    do_submit(transfer_client)



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

ACCESS_POINT_ID = str(uuid.uuid4())


RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="transfer",
        method="GET",
        path=f"/v2/stream_access_points/{ACCESS_POINT_ID}",
        json={
            "data": {
                "attributes": {
                    "advertised_owner": "john@globus.org",
                    "contact_email": None,
                    "contact_info": None,
                    "department": None,
                    "description": None,
                    "display_name": "Buzz Dev Listener",
                    "info_link": None,
                    "keywords": None,
                    "organization": None,
                    "tlsftp_server": (
                        "tlsftp://s-463c7.e7d5e.8540."
                        "test3.zones.dnsteam.globuscs.info:443"
                    ),
                },
                "id": ACCESS_POINT_ID,
                "relationships": {
                    "host_endpoint": {
                        "data": {
                            "id": "d6428474-c308-4a2d-8a86-d377915d978b",
                            "type": "Endpoint",
                        }
                    }
                },
                "type": "StreamAccessPoint",
            },
            "meta": {"request_id": "55QRq2iBa"},
        },
        metadata={
            "access_point_id": ACCESS_POINT_ID,
        },
    ),
)



from __future__ import annotations

import logging
import typing as t
import uuid

from globus_sdk import exc
from globus_sdk._internal.remarshal import strseq_iter, strseq_listify
from globus_sdk._missing import MISSING, MissingType
from globus_sdk.authorizers import BasicAuthorizer
from globus_sdk.response import GlobusHTTPResponse
from globus_sdk.scopes import Scope, ScopeParser
from globus_sdk.transport import RequestsTransport, RetryConfig

from ..flow_managers import GlobusAuthorizationCodeFlowManager
from ..response import OAuthClientCredentialsResponse, OAuthDependentTokenResponse
from .base_login_client import AuthLoginClient

log = logging.getLogger(__name__)


class ConfidentialAppAuthClient(AuthLoginClient):
    """
    This is a specialized type of ``AuthLoginClient`` used to represent an App with
    a Client ID and Client Secret wishing to communicate with Globus Auth.
    It must be given a Client ID and a Client Secret, and furthermore, these
    will be used to establish a :class:`BasicAuthorizer <globus_sdk.BasicAuthorizer>`
    for authorization purposes.
    Additionally, the Client ID is stored for use in various calls.

    Confidential Applications are those which have their own credentials for
    authenticating against Globus Auth.

    :param client_id: The ID of the application provided by registration with
        Globus Auth.
    :param client_secret: The secret string to use for authentication. Secrets can be
        generated via the Globus developers interface.

    All other initialization parameters are passed through to ``BaseClient``.

    .. automethodlist:: globus_sdk.ConfidentialAppAuthClient
    """

    def __init__(
        self,
        client_id: uuid.UUID | str,
        client_secret: str,
        environment: str | None = None,
        base_url: str | None = None,
        app_name: str | None = None,
        transport: RequestsTransport | None = None,
        retry_config: RetryConfig | None = None,
    ) -> None:
        super().__init__(
            client_id=client_id,
            authorizer=BasicAuthorizer(str(client_id), client_secret),
            environment=environment,
            base_url=base_url,
            app_name=app_name,
            transport=transport,
            retry_config=retry_config,
        )

    def oauth2_client_credentials_tokens(
        self, requested_scopes: str | Scope | t.Iterable[str | Scope]
    ) -> OAuthClientCredentialsResponse:
        r"""
        Perform an OAuth2 Client Credentials Grant to get access tokens which
        directly represent your client and allow it to act on its own
        (independent of any user authorization).
        This method does not use a ``GlobusOAuthFlowManager`` because it is not
        at all necessary to do so.

        :param requested_scopes: The scopes on the token(s) being requested.

        For example, with a Client ID of "CID1001" and a Client Secret of
        "RAND2002", you could use this grant type like so:

        .. code-block:: pycon

            >>> client = ConfidentialAppAuthClient("CID1001", "RAND2002")
            >>> tokens = client.oauth2_client_credentials_tokens(
            ...     "urn:globus:auth:scope:transfer.api.globus.org:all"
            ... )
            >>> transfer_token_info = tokens.by_resource_server["transfer.api.globus.org"]
            >>> transfer_token = transfer_token_info["access_token"]
        """  # noqa: E501
        requested_scopes_string = ScopeParser.serialize(requested_scopes)
        log.debug(
            "Fetching token(s) using client credentials, "
            f"scope={requested_scopes_string}"
        )
        return self.oauth2_token(
            {"grant_type": "client_credentials", "scope": requested_scopes_string},
            response_class=OAuthClientCredentialsResponse,
        )

    def oauth2_start_flow(
        self,
        redirect_uri: str,
        requested_scopes: str | Scope | t.Iterable[str | Scope],
        *,
        state: str = "_default",
        refresh_tokens: bool = False,
    ) -> GlobusAuthorizationCodeFlowManager:
        """
        Starts or resumes an Authorization Code OAuth2 flow.

        Under the hood, this is done by instantiating a
        :class:`GlobusAuthorizationCodeFlowManager
        <.GlobusAuthorizationCodeFlowManager>`

        :param redirect_uri: The page that users should be directed to after
            authenticating at the authorize URL.
        :param requested_scopes: The scopes on the token(s) being requested.
        :param state: This string allows an application to pass information back to
            itself in the course of the OAuth flow. Because the user will navigate away
            from the application to complete the flow, this parameter lets the app pass
            an arbitrary string from the starting page to the ``redirect_uri``
        :param refresh_tokens: When True, request refresh tokens in addition to access
            tokens. [Default: ``False``]

        .. tab-set::

            .. tab-item:: Example Usage

                You can see an example of this flow :ref:`in the usage examples
                <examples_three_legged_oauth_login>`.

            .. tab-item:: API Info

                The Authorization Code Grant flow is described
                `in the Globus Auth Specification
                <https://docs.globus.org/api/auth/developer-guide/#obtaining-authorization>`_.
        """
        log.debug("Starting OAuth2 Authorization Code Grant Flow")
        self.current_oauth2_flow_manager = GlobusAuthorizationCodeFlowManager(
            self,
            redirect_uri,
            requested_scopes=requested_scopes,
            state=state,
            refresh_tokens=refresh_tokens,
        )
        return self.current_oauth2_flow_manager

    def oauth2_get_dependent_tokens(
        self,
        token: str,
        *,
        refresh_tokens: bool = False,
        scope: str | t.Iterable[str] | MissingType = MISSING,
        additional_params: dict[str, t.Any] | None = None,
    ) -> OAuthDependentTokenResponse:
        """
        Fetch Dependent Tokens from Globus Auth.

        This exchanges a token given to this client for a new set of tokens
        which give it access to resource servers on which it depends.
        This grant type is intended for use by Resource Servers playing out the
        following scenario:

          1. User has tokens for Service A, but Service A requires access to
             Service B on behalf of the user
          2. Service B should not see tokens scoped for Service A
          3. Service A therefore requests tokens scoped only for Service B,
             based on tokens which were originally scoped for Service A...

        In order to do this exchange, the tokens for Service A must have scopes
        which depend on scopes for Service B (the services' scopes must encode
        their relationship). As long as that is the case, Service A can use
        this Grant to get those "Dependent" or "Downstream" tokens for Service B.

        :param token: An access token as a string
        :param refresh_tokens: When True, request dependent refresh tokens in addition
            to access tokens. [Default: ``False``]
        :param scope: The scope or scopes of the dependent tokens which are being
            requested. Applications are recommended to provide this string to ensure
            that they are receiving the tokens they expect. If omitted, all available
            dependent tokens will be returned.
        :param additional_params: Additional parameters to include in the request body

        .. tab-set::

            .. tab-item:: Example Usage

                Given a token, getting a dependent token for Globus Groups might
                look like the following:

                .. code-block:: python

                    ac = globus_sdk.ConfidentialAppAuthClient(CLIENT_ID, CLIENT_SECRET)
                    dependent_token_data = ac.oauth2_get_dependent_tokens(
                        "<token_string>",
                        scope="urn:globus:auth:scope:groups.api.globus.org:view_my_groups_and_memberships",
                    )

                    group_token_data = dependent_token_data.by_resource_server["groups.api.globus.org"]
                    group_token = group_token_data["access_token"]

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.oauth2_get_dependent_tokens
                    :case: groups

            .. tab-item:: API Info

                ``POST /v2/oauth2/token``

                .. extdoclink:: Dependent Token Grant
                    :ref: auth/reference/##dependent_token_grant_post_v2oauth2token
        """  # noqa: E501
        log.debug(
            "Getting dependent tokens from access token"
            f"additional_params={additional_params}"
        )
        form_data = {
            "grant_type": "urn:globus:auth:grant_type:dependent_token",
            "token": token,
            # the internal parameter is 'access_type', but using the name
            # 'refresh_tokens' is consistent with the rest of the SDK and better
            # communicates expectations back to the user than the OAuth2 spec wording
            "access_type": "offline" if refresh_tokens else MISSING,
            "scope": (" ".join(strseq_iter(scope)) if scope is not MISSING else scope),
            **(additional_params or {}),
        }
        return self.oauth2_token(form_data, response_class=OAuthDependentTokenResponse)

    def oauth2_token_introspect(
        self,
        token: str,
        *,
        include: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> GlobusHTTPResponse:
        """
        Get information about a Globus Auth token.

        :param token: An Access Token as a raw string, being evaluated
        :param include: A value for the ``include`` parameter in the request body.
            Default is to omit the parameter.
        :param query_params: Any additional parameters to be passed through
            as query params.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    ac = globus_sdk.ConfidentialAppAuthClient(CLIENT_ID, CLIENT_SECRET)
                    ac.oauth2_token_introspect("<token_string>")

                Get information about a Globus Auth token including the full identity
                set of the user to whom it belongs

                .. code-block:: python

                    ac = globus_sdk.ConfidentialAppAuthClient(CLIENT_ID, CLIENT_SECRET)
                    data = ac.oauth2_token_introspect("<token_string>", include="identity_set")
                    for identity in data["identity_set"]:
                        print('token authenticates for "{}"'.format(identity))

            .. tab-item:: API Info

                ``POST /v2/oauth2/token/introspect``

                .. extdoclink:: Token Introspection
                    :ref: auth/reference/#token_introspection_post_v2_oauth2_token_introspect
        """  # noqa: E501
        log.debug("Checking token validity (introspect)")
        body = {
            "token": token,
            "include": include,
        }
        return self.post(
            "/v2/oauth2/token/introspect",
            data=body,
            encoding="form",
            query_params=query_params,
        )

    def create_child_client(
        self,
        name: str,
        *,
        public_client: bool | MissingType = MISSING,
        client_type: (
            t.Literal[
                "client_identity",
                "confidential_client",
                "globus_connect_server",
                "public_installed_client",
                "hybrid_confidential_client_resource_server",
                "resource_server",
            ]
            | MissingType
        ) = MISSING,
        visibility: t.Literal["public", "private"] | MissingType = MISSING,
        redirect_uris: t.Iterable[str] | MissingType = MISSING,
        terms_and_conditions: str | MissingType = MISSING,
        privacy_policy: str | MissingType = MISSING,
        required_idp: uuid.UUID | str | MissingType = MISSING,
        preselect_idp: uuid.UUID | str | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> GlobusHTTPResponse:
        """
        Create a new client. Requires the ``manage_projects`` scope.

        :param name: The display name shown to users on consents. May not contain
            linebreaks.
        :param public_client: This is used to infer which OAuth grant_types the client
            will be able to use. Should be false if the client is capable of keeping
            secret credentials (such as clients running on a server) and true if it is
            not (such as native apps). After creation this value is immutable. This
            option is mutually exclusive with ``client_type``, exactly one must be
            given.
        :param client_type: Defines the type of client that will be created. This
            option is mutually exclusive with ``public_client``, exactly one must
            be given.

            .. dropdown:: Values for ``client_type``

                .. list-table::

                    * - ``"confidential_client"``
                      - Applications that are OAuth confidential clients, and can
                        manage a client secret and requests for user consent.
                    * - ``"public_installed_client"``
                      - Applications that are OAuth public clients or native
                        applications that are distributed to users, and thus cannot
                        manage a client secret.
                    * - ``"client_identity"``
                      - Applications that authenticate and act as the application
                        itself. These applications are used for automation and as
                        service or community accounts, and do NOT act on behalf of
                        other users. Also known as a "Service Account".
                    * - ``"resource_server"``
                      - An API (OAuth resource server) that uses Globus Auth tokens for
                        authentication. Users accessing the service login via Globus and
                        consent for the client to use your API.
                    * - ``"globus_connect_server"``
                      - Create a client that will service as a Globus Connect Server
                        endpoint.
                    * - ``"hybrid_confidential_client_resource_server"``
                      - A client which can use any behavior with Globus Auth - an
                        application (confidential or public client), service account,
                        or API.

        :param visibility: If set to "public", any authenticated entity can view it.
            When set to "private", only entities in the same project as the client can
            view it.
        :param redirect_uris: list of URIs that may be used in OAuth authorization
            flows.
        :param terms_and_conditions: URL of client's terms and conditions.
        :param privacy_policy: URL of client's privacy policy.
        :param required_idp: In order to use this client a user must have an identity
            from this IdP in their identity set.
        :param preselect_idp: This preselects the given IdP on the Globus Auth login
            page if the user is not already authenticated.
        :param additional_fields: Any additional parameters to be passed through.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> project_id = ...
                    >>> r = ac.create_child_client(
                    ...     "My client",
                    ...     True,
                    ...     True,
                    ... )
                    >>> client_id = r["client"]["id"]

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.create_child_client

            .. tab-item:: API Info

                ``POST /v2/api/clients``

                .. extdoclink:: Create Client
                    :ref: auth/reference/#create_client
        """
        # Must specify exactly one of public_client or client_type
        if public_client is not MISSING and client_type is not MISSING:
            raise exc.GlobusSDKUsageError(
                "AuthClient.create_client does not take both "
                "'public_client' and 'client_type'. These are mutually exclusive."
            )
        if public_client is MISSING and client_type is MISSING:
            raise exc.GlobusSDKUsageError(
                "AuthClient.create_client requires either 'public_client' or "
                "'client_type'."
            )
        # terms_and_conditions and privacy_policy must both be set or unset
        if bool(terms_and_conditions) ^ bool(privacy_policy):
            raise exc.GlobusSDKUsageError(
                "terms_and_conditions and privacy_policy must both be set or unset"
            )
        links: dict[str, str | MissingType] | MissingType = MISSING
        if terms_and_conditions and privacy_policy:
            links = {
                "terms_and_conditions": terms_and_conditions,
                "privacy_policy": privacy_policy,
            }

        body: dict[str, t.Any] = {
            "name": name,
            "visibility": visibility,
            "required_idp": required_idp,
            "preselect_idp": preselect_idp,
            "public_client": public_client,
            "client_type": client_type,
            "redirect_uris": strseq_listify(redirect_uris),
            "links": links,
            **(additional_fields or {}),
        }

        return self.post("/v2/api/clients", data={"client": body})



import uuid

import globus_sdk
from globus_sdk.testing import load_response

ENDPOINT_CONFIG = """
display_name: My Endpoint
engine:
  type: GlobusComputeEngine
"""


def test_register_endpoint(compute_client_v2: globus_sdk.ComputeClientV2):
    meta = load_response(compute_client_v2.register_endpoint).metadata
    register_doc = {
        "endpoint_uuid": meta["endpoint_id"],
        "endpoint_name": "my-endpoint",
        "display_name": "My Endpoint",
        "version": "2.31.0",
        "multi_user": False,
        "allowed_functions": [str(uuid.uuid1())],
        "authentication_policy": str(uuid.uuid1()),
        "metadata": {
            "endpoint_config": ENDPOINT_CONFIG.strip(),
            "user_config_template": "",
            "user_config_schema": {},
            "description": "My endpoint description",
            "ip_address": "140.221.112.13",
            "hostname": "my-hostname",
            "local_user": "user1",
            "sdk_version": "2.31.0",
            "endpoint_version": "2.31.0",
        },
    }

    res = compute_client_v2.register_endpoint(data=register_doc)

    assert res.http_status == 200
    assert res.data["endpoint_id"] == meta["endpoint_id"]



from globus_sdk.testing import get_last_request, load_response


def test_list_bookmarks(client):
    meta = load_response(client.list_bookmarks).metadata

    res = client.list_bookmarks(
        query_params={"include": "collection"},
    )

    assert res.http_status == 200
    for i, bm in enumerate(res["data"]):
        assert bm["type"] == "Bookmark"
        assert bm["id"] == meta["bookmarks"][i]["bookmark_id"]
        assert bm["attributes"]["name"] == meta["bookmarks"][i]["name"]
        assert bm["attributes"]["path"] == meta["bookmarks"][i]["path"]
        assert bm["attributes"]["pinned"] == meta["bookmarks"][i]["pinned"]

        assert (
            bm["relationships"]["collection"]["data"]["id"]
            == meta["bookmarks"][i]["collection_id"]
        )

    for i, col in enumerate(res["included"]):
        assert col["type"] == "Collection"
        assert col["id"] == meta["bookmarks"][i]["collection_id"]

    req = get_last_request()
    assert "include=collection" in req.url
    assert req.body is None



import inspect
import sys
import types
import typing as t
import uuid

import pytest

from globus_sdk import (
    CollectionDocument,
    GoogleCloudStorageCollectionPolicies,
    GuestCollectionDocument,
    MappedCollectionDocument,
    POSIXCollectionPolicies,
    POSIXStagingCollectionPolicies,
)
from globus_sdk._missing import MISSING, MissingType, filter_missing
from globus_sdk.transport import JSONRequestEncoder

if sys.version_info >= (3, 10):
    UnionTypes = (t.Union, types.UnionType)
else:
    UnionTypes = (t.Union,)

STUB_SG_ID = uuid.uuid1()  # storage gateway
STUB_MC_ID = uuid.uuid1()  # mapped collection
STUB_UC_ID = uuid.uuid1()  # user credential


MappedCollectionSignature = inspect.signature(MappedCollectionDocument.__init__)
GuestCollectionSignature = inspect.signature(GuestCollectionDocument.__init__)


def test_collection_base_abstract():
    with pytest.raises(TypeError):
        CollectionDocument()


def test_collection_type_field():
    m = MappedCollectionDocument(
        storage_gateway_id=STUB_SG_ID, collection_base_path="/"
    )
    g = GuestCollectionDocument(
        mapped_collection_id=STUB_MC_ID,
        user_credential_id=STUB_UC_ID,
        collection_base_path="/",
    )
    assert m["collection_type"] == "mapped"
    assert g["collection_type"] == "guest"


@pytest.mark.parametrize(
    "use_kwargs,doc_version",
    [
        ({}, "1.0.0"),
        ({"user_message_link": "https://example.net/"}, "1.1.0"),
        ({"user_message": "kthxbye"}, "1.1.0"),
        ({"user_message": ""}, "1.1.0"),
        ({"enable_https": True}, "1.1.0"),
        ({"enable_https": False}, "1.1.0"),
        # a string of length > 64
        ({"user_message": "long message..." + "x" * 100}, "1.7.0"),
    ],
)
def test_datatype_version_deduction(use_kwargs, doc_version):
    m = MappedCollectionDocument(**use_kwargs)
    assert m["DATA_TYPE"] == f"collection#{doc_version}"
    g = GuestCollectionDocument(**use_kwargs)
    assert g["DATA_TYPE"] == f"collection#{doc_version}"


@pytest.mark.parametrize(
    "use_kwargs,doc_version",
    [
        ({"sharing_users_allow": "sirosen"}, "1.2.0"),
        ({"sharing_users_allow": ["sirosen", "aaschaer"]}, "1.2.0"),
        ({"sharing_users_deny": "sirosen"}, "1.2.0"),
        ({"sharing_users_deny": ["sirosen", "aaschaer"]}, "1.2.0"),
        ({"force_verify": True}, "1.4.0"),
        ({"force_verify": False}, "1.4.0"),
        ({"disable_anonymous_writes": True}, "1.5.0"),
        ({"disable_anonymous_writes": False}, "1.5.0"),
        ({"guest_auth_policy_id": str(uuid.uuid4())}, "1.6.0"),
        ({"delete_protected": False}, "1.8.0"),
        # combining a long user_message (which uses callback-based detection) with
        # higher and lower bounding fields needs to apply correctly
        (
            {"force_verify": False, "user_message": "long message..." + "x" * 100},
            "1.7.0",
        ),
        (
            {"delete_protected": False, "user_message": "long message..." + "x" * 100},
            "1.8.0",
        ),
    ],
)
def test_datatype_version_deduction_mapped_specific_fields(use_kwargs, doc_version):
    d = MappedCollectionDocument(**use_kwargs)
    assert d["DATA_TYPE"] == f"collection#{doc_version}"


def test_datatype_version_deduction_add_custom(monkeypatch):
    custom_field = "foo-made-up-field"
    monkeypatch.setitem(
        CollectionDocument.DATATYPE_VERSION_IMPLICATIONS, custom_field, (1, 20, 0)
    )

    m = MappedCollectionDocument(
        storage_gateway_id=STUB_SG_ID,
        collection_base_path="/",
        additional_fields={custom_field: "foo"},
    )
    assert m["DATA_TYPE"] == "collection#1.20.0"
    g = GuestCollectionDocument(
        mapped_collection_id=STUB_MC_ID,
        user_credential_id=STUB_UC_ID,
        collection_base_path="/",
        additional_fields={custom_field: "foo"},
    )
    assert g["DATA_TYPE"] == "collection#1.20.0"


@pytest.mark.parametrize(
    "policies_type",
    (
        dict,
        POSIXCollectionPolicies,
        POSIXStagingCollectionPolicies,
        GoogleCloudStorageCollectionPolicies,
    ),
)
def test_collection_policies_field_encoded(policies_type):
    if policies_type is dict:
        policy_data = {"spam": "eggs"}
    elif policies_type in (POSIXCollectionPolicies, POSIXStagingCollectionPolicies):
        policy_data = policies_type(
            sharing_groups_allow=["foo", "bar"],
            sharing_groups_deny="baz",
            additional_fields={"spam": "eggs"},
        )
    elif policies_type is GoogleCloudStorageCollectionPolicies:
        policy_data = GoogleCloudStorageCollectionPolicies(
            project="foo", additional_fields={"spam": "eggs"}
        )
    else:
        raise NotImplementedError

    # only Mapped Collections support a policies subdocument
    doc = MappedCollectionDocument(
        storage_gateway_id=STUB_SG_ID,
        collection_base_path="/",
        policies=policy_data,
    )

    assert "policies" in doc
    assert isinstance(doc["policies"], policies_type)

    encoder = JSONRequestEncoder()
    request_data = encoder.encode("POST", "bogus.url.example", {}, doc, {}).json

    if policies_type is dict:
        assert request_data["policies"] == {"spam": "eggs"}
    elif policies_type is POSIXCollectionPolicies:
        assert request_data["policies"] == {
            "DATA_TYPE": "posix_collection_policies#1.0.0",
            "spam": "eggs",
            "sharing_groups_allow": ["foo", "bar"],
            "sharing_groups_deny": ["baz"],
        }
    elif policies_type is POSIXStagingCollectionPolicies:
        assert request_data["policies"] == {
            "DATA_TYPE": "posix_staging_collection_policies#1.0.0",
            "spam": "eggs",
            "sharing_groups_allow": ["foo", "bar"],
            "sharing_groups_deny": ["baz"],
        }
    elif policies_type is GoogleCloudStorageCollectionPolicies:
        assert request_data["policies"] == {
            "DATA_TYPE": "google_cloud_storage_collection_policies#1.0.0",
            "spam": "eggs",
            "project": "foo",
        }
    else:
        raise NotImplementedError


# these test cases enumerate parameters for Guest Collections and Mapped Collections
# and ensure that they're defined on one class but not the other
@pytest.mark.parametrize(
    "fieldname",
    (
        "allow_guest_collections",
        "delete_protected",
        "disable_anonymous_writes",
        "domain_name",
        "guest_auth_policy_id",
        "policies",
        "sharing_restrict_paths",
        "sharing_users_allow",
        "sharing_users_deny",
        "storage_gateway_id",
    ),
)
def test_settings_which_are_only_supported_in_mapped_collections(fieldname):
    assert fieldname in MappedCollectionSignature.parameters
    assert fieldname not in GuestCollectionSignature.parameters


@pytest.mark.parametrize(
    "fieldname",
    (
        "mapped_collection_id",
        "user_credential_id",
    ),
)
def test_settings_which_are_only_supported_in_guest_collections(fieldname):
    assert fieldname in GuestCollectionSignature.parameters
    assert fieldname not in MappedCollectionSignature.parameters


@pytest.mark.parametrize(
    "fieldname",
    (
        "disable_verify",
        "enable_https",
        "force_encryption",
        "force_verify",
        "public",
        "delete_protected",
        "allow_guest_collections",
        "disable_anonymous_writes",
    ),
)
@pytest.mark.parametrize("value", (True, False, None))
def test_mapped_collection_opt_bool(fieldname, value):
    data = {}
    if value is not None:
        data[fieldname] = value

    doc = filter_missing(
        MappedCollectionDocument(
            storage_gateway_id=STUB_SG_ID,
            collection_base_path="/",
            **data,
        )
    )
    if value is not None:
        assert fieldname in doc
        assert doc[fieldname] == value
    else:
        assert fieldname not in doc


common_collection_fields = [
    ("collection_base_path", (str, MissingType)),
    ("contact_email", (str, None, MissingType)),
    ("contact_info", (str, None, MissingType)),
    ("default_directory", (str, MissingType)),
    ("department", (str, None, MissingType)),
    ("description", (str, None, MissingType)),
    ("display_name", (str, MissingType)),
    ("identity_id", (t.Union[uuid.UUID, str], MissingType)),
    ("info_link", (str, None, MissingType)),
    ("organization", (str, MissingType)),
    ("user_message", (str, None, MissingType)),
    ("user_message_link", (str, None, MissingType)),
    ("keywords", (t.Iterable[str], MissingType)),
    ("disable_verify", (bool, MissingType)),
    ("enable_https", (bool, MissingType)),
    ("force_encryption", (bool, MissingType)),
    ("force_verify", (bool, MissingType)),
    ("public", (bool, MissingType)),
]


mapped_collection_fields = [
    *common_collection_fields,
    ("domain_name", (str, MissingType)),
    ("guest_auth_policy_id", (t.Union[uuid.UUID, str], None, MissingType)),
    ("disable_anonymous_writes", (bool, MissingType)),
    ("policies", (t.Dict[str, t.Any], MissingType)),
]


guest_collection_fields = [
    *common_collection_fields,
    ("mapped_collection_id", (t.Union[uuid.UUID, str], MissingType)),
    ("user_credential_id", (t.Union[uuid.UUID, str], MissingType)),
    ("activity_notification_policy", (t.Dict[str, t.List[str]], MissingType)),
]


def expand_collection_fields(fields):
    """
    Expand each collection field into (field, valid_value)
    """
    return [
        (param, value)
        for param, types in fields
        for _type in types
        for value in _gen_value(_type)
    ]


def _gen_value(_type):
    """
    Return a list of valid values for type _type.
    """
    if _type is MissingType:
        return [MISSING]
    if _type is None:
        return [None]
    if _type is str:
        return ["STRING"]
    if _type is bool:
        return [True, False]
    # Union[UUID, str] | (UUID | str)
    if t.get_origin(_type) in UnionTypes and set(t.get_args(_type)) == {uuid.UUID, str}:
        return [str(uuid.uuid1()), uuid.uuid1()]
    if _type is t.Iterable[str]:
        return [[], ["a", "b", "c"]]
    if _type is t.Dict[str, t.Any]:
        return [{"A": 1}]
    if _type is t.Dict[str, t.List[str]]:
        return [{"a": ["b", "c"]}]

    raise AssertionError(f"Unexpected Type: {_type}")


@pytest.mark.parametrize(
    "fieldname,value",
    expand_collection_fields(mapped_collection_fields),
)
def test_mapped_collection_fields(fieldname, value):
    """
    Verify that each field in the mapped collection document can be set to a valid
    value.
    """
    data = {}
    if value != MISSING:
        data[fieldname] = value

    doc = MappedCollectionDocument(**data)
    assert doc[fieldname] == value


@pytest.mark.parametrize(
    "fieldname,value",
    expand_collection_fields(guest_collection_fields),
)
def test_guest_collection_fields(fieldname, value):
    """
    Verify that each field in the guest collection document can be set to a valid
    value.
    """
    data = {}
    if value != MISSING:
        data[fieldname] = value

    doc = GuestCollectionDocument(**data)
    assert doc[fieldname] == value


# regression test for a typo which caused this to be set improperly to the wrong key
@pytest.mark.parametrize("value", ("inbound", "outbound", "all"))
@pytest.mark.parametrize("collection_type", ("mapped", "guest"))
def test_can_set_restrict_transfers_to_high_assurance(value, collection_type):
    if collection_type == "mapped":
        c = MappedCollectionDocument(
            storage_gateway_id=STUB_SG_ID,
            collection_base_path="/",
            restrict_transfers_to_high_assurance=value,
        )
    else:
        c = GuestCollectionDocument(
            mapped_collection_id=STUB_MC_ID,
            user_credential_id=STUB_UC_ID,
            collection_base_path="/",
            restrict_transfers_to_high_assurance=value,
        )

    assert c["restrict_transfers_to_high_assurance"] == value



"""
Test that globus_sdk.testing can accept a non-default requests mock
"""

import pytest
import requests
import responses

from globus_sdk import GlobusHTTPResponse, GroupsClient
from globus_sdk.testing import get_last_request, load_response


@pytest.fixture
def custom_requests_mock():
    responses.stop()

    with responses.RequestsMock() as m:
        yield m


def test_get_last_request_on_empty_custom_mock_returns_none(custom_requests_mock):
    assert get_last_request(requests_mock=custom_requests_mock) is None


def test_get_last_request_on_custom_mock_populated_via_manual_load(
    custom_requests_mock,
):
    custom_requests_mock.add("GET", "https://example.org/", json={"foo": "bar"})
    r = requests.get("https://example.org/")
    assert r.json() == {"foo": "bar"}

    req = get_last_request(requests_mock=custom_requests_mock)
    assert req is not None
    assert req.body is None  # "example" assertion about a request


def test_register_client_method_and_call_on_custom_mock(custom_requests_mock):
    gc = GroupsClient()
    loaded = load_response(gc.get_group, requests_mock=custom_requests_mock)

    group_id = loaded.metadata["group_id"]
    data = gc.get_group(group_id)

    assert isinstance(data, GlobusHTTPResponse)



import datetime

import pytest

from globus_sdk import MISSING, DeleteData, TransferData
from globus_sdk.services.transfer.client import _format_filter
from tests.common import GO_EP1_ID, GO_EP2_ID


def test_transfer_init_no_params():
    """
    Creates a TransferData object without optional parameters and
    verifies field initialization.
    """
    # default init
    tdata = TransferData(GO_EP1_ID, GO_EP2_ID)
    assert tdata["DATA_TYPE"] == "transfer"
    assert tdata["source_endpoint"] == GO_EP1_ID
    assert tdata["destination_endpoint"] == GO_EP2_ID
    assert tdata["submission_id"] is MISSING
    assert "DATA" in tdata
    assert len(tdata["DATA"]) == 0


def test_transfer_init_w_params():
    # init with params
    label = "label"
    params = {"param1": "value1", "param2": "value2"}
    tdata = TransferData(
        GO_EP1_ID,
        GO_EP2_ID,
        label=label,
        sync_level="exists",
        additional_fields=params,
    )
    assert tdata["label"] == label
    assert tdata["submission_id"] is MISSING
    # sync_level of "exists" should be converted to 0
    assert tdata["sync_level"] == 0
    for par in params:
        assert tdata[par] == params[par]


def test_transfer_add_item():
    """
    Adds items to TransferData, verifies results
    """
    tdata = TransferData(GO_EP1_ID, GO_EP2_ID)
    # add item
    source_path = "source/path/"
    dest_path = "dest/path/"
    tdata.add_item(source_path, dest_path)
    # verify results
    assert len(tdata["DATA"]) == 1
    data = tdata["DATA"][0]
    assert data["DATA_TYPE"] == "transfer_item"
    assert data["source_path"] == source_path
    assert data["destination_path"] == dest_path
    assert data["recursive"] == MISSING
    assert data["external_checksum"] == MISSING
    assert data["checksum_algorithm"] == MISSING

    # add recursive item
    tdata.add_item(source_path, dest_path, recursive=True)
    # verify results
    assert len(tdata["DATA"]) == 2
    r_data = tdata["DATA"][1]
    assert r_data["DATA_TYPE"] == "transfer_item"
    assert r_data["source_path"] == source_path
    assert r_data["destination_path"] == dest_path
    assert r_data["recursive"]
    assert data["external_checksum"] == MISSING
    assert data["checksum_algorithm"] == MISSING

    # item with checksum
    checksum = "d577273ff885c3f84dadb8578bb41399"
    algorithm = "MD5"
    tdata.add_item(
        source_path, dest_path, external_checksum=checksum, checksum_algorithm=algorithm
    )
    assert len(tdata["DATA"]) == 3
    c_data = tdata["DATA"][2]
    assert c_data["DATA_TYPE"] == "transfer_item"
    assert c_data["source_path"] == source_path
    assert c_data["destination_path"] == dest_path
    assert c_data["recursive"] == MISSING
    assert c_data["external_checksum"] == checksum
    assert c_data["checksum_algorithm"] == algorithm

    # add an item which uses `additional_fields`
    addfields = {"foo": "bar", "bar": "baz"}
    tdata.add_item(source_path, dest_path, additional_fields=addfields)
    assert len(tdata["DATA"]) == 4
    fields_data = tdata["DATA"][3]
    assert fields_data["DATA_TYPE"] == "transfer_item"
    assert fields_data["source_path"] == source_path
    assert fields_data["destination_path"] == dest_path
    assert fields_data["recursive"] == MISSING
    assert all(fields_data[k] == v for k, v in addfields.items())


@pytest.mark.parametrize(
    "args, kwargs",
    (
        ((GO_EP1_ID,), {}),
        ((), {"endpoint": GO_EP1_ID}),
    ),
)
def test_delete_data_noparams_init(args, kwargs):
    # the minimal, required argument is the endpoint ID -- less than that results
    # in a TypeError because the signature is not obeyed
    ddata = DeleteData(*args, **kwargs)
    assert ddata["DATA_TYPE"] == "delete"
    assert ddata["endpoint"] == GO_EP1_ID
    assert ddata["submission_id"] is MISSING
    assert "DATA" in ddata
    assert len(ddata["DATA"]) == 0


@pytest.mark.parametrize(
    "add_kwargs",
    (
        {"recursive": True},
        {"ignore_missing": True},
        {"interpret_globs": True},
        {"label": "somelabel"},
    ),
)
def test_delete_init_with_supported_parameters(add_kwargs):
    ddata = DeleteData(GO_EP1_ID, **add_kwargs)
    for k, v in add_kwargs.items():
        assert ddata[k] == v


def test_delete_init_with_additional_fields():
    params = {"param1": "value1", "param2": "value2"}
    ddata = DeleteData(GO_EP1_ID, additional_fields=params)
    assert ddata["param1"] == "value1"
    assert ddata["param2"] == "value2"


def test_delete_add_item():
    """
    Adds items to DeleteData, verifies results
    """
    ddata = DeleteData(GO_EP1_ID)

    # add normal item
    path = "source/path/"
    ddata.add_item(path)
    assert len(ddata["DATA"]) == 1
    data = ddata["DATA"][0]
    assert data["DATA_TYPE"] == "delete_item"
    assert data["path"] == path

    # add an item which uses `additional_fields`
    addfields = {"foo": "bar", "bar": "baz"}
    ddata.add_item(path, additional_fields=addfields)
    assert len(ddata["DATA"]) == 2
    fields_data = ddata["DATA"][1]
    assert fields_data["DATA_TYPE"] == "delete_item"
    assert fields_data["path"] == path
    assert all(fields_data[k] == v for k, v in addfields.items())


def test_delete_iter_items():
    ddata = DeleteData(GO_EP1_ID)
    # add item
    ddata.add_item("abc/")
    ddata.add_item("def/")

    # order preserved, well-formed
    as_list = list(ddata.iter_items())
    assert len(as_list) == 2

    def check_item(x, path):
        assert isinstance(x, dict)
        assert x.get("DATA_TYPE") == "delete_item"
        assert x.get("path") == path

    check_item(as_list[0], "abc/")
    check_item(as_list[1], "def/")


def test_transfer_iter_items():
    tdata = TransferData(GO_EP1_ID, GO_EP2_ID)
    tdata.add_item("source/abc.txt", "dest/abc.txt")
    tdata.add_item("source/def/", "dest/def/", recursive=True)

    # order preserved, well-formed
    as_list = list(tdata.iter_items())
    assert len(as_list) == 2

    def check_item(x, src, dst, params=None):
        params = params or {}
        assert isinstance(x, dict)
        assert x.get("DATA_TYPE") == "transfer_item"
        assert x.get("source_path") == src
        assert x.get("destination_path") == dst
        for k, v in params.items():
            assert x.get(k) == v

    check_item(as_list[0], "source/abc.txt", "dest/abc.txt")
    check_item(as_list[1], "source/def/", "dest/def/", {"recursive": True})


@pytest.mark.parametrize("n_succeeded", [None, True, False])
@pytest.mark.parametrize("n_failed", [None, True, False])
@pytest.mark.parametrize("n_inactive", [None, True, False])
def test_notification_options(n_succeeded, n_failed, n_inactive):
    notify_kwargs = {}
    if n_succeeded is not None:
        notify_kwargs["notify_on_succeeded"] = n_succeeded
    if n_failed is not None:
        notify_kwargs["notify_on_failed"] = n_failed
    if n_inactive is not None:
        notify_kwargs["notify_on_inactive"] = n_inactive

    ddata = DeleteData(GO_EP1_ID, **notify_kwargs)
    tdata = TransferData(GO_EP1_ID, GO_EP2_ID, **notify_kwargs)

    def _default(x):
        return x if x is not None else True

    expect = {
        "notify_on_succeeded": _default(n_succeeded),
        "notify_on_failed": _default(n_failed),
        "notify_on_inactive": _default(n_inactive),
    }
    for k, v in expect.items():
        if k in notify_kwargs:
            assert tdata[k] is v
            assert ddata[k] is v
        else:
            assert tdata[k] is MISSING
            assert ddata[k] is MISSING


@pytest.mark.parametrize(
    "sync_level, result",
    [
        ("exists", 0),
        (0, 0),
        ("size", 1),
        (1, 1),
        ("mtime", 2),
        (2, 2),
        ("checksum", 3),
        (3, 3),
        ("EXISTS", ValueError),
        ("hash", ValueError),
        (100, 100),
    ],
)
def test_transfer_sync_levels_result(sync_level, result):
    if isinstance(result, type) and issubclass(result, Exception):
        with pytest.raises(result):
            TransferData(GO_EP1_ID, GO_EP2_ID, sync_level=sync_level)
    else:
        tdata = TransferData(GO_EP1_ID, GO_EP2_ID, sync_level=sync_level)
        assert tdata["sync_level"] == result


def test_add_filter_rule():
    tdata = TransferData(GO_EP1_ID, GO_EP2_ID)
    assert "filter_rules" not in tdata

    tdata.add_filter_rule("*.tgz", type="file")
    assert "filter_rules" in tdata
    assert isinstance(tdata["filter_rules"], list)
    assert len(tdata["filter_rules"]) == 1
    assert tdata["filter_rules"][0]["DATA_TYPE"] == "filter_rule"
    assert tdata["filter_rules"][0]["method"] == "exclude"
    assert tdata["filter_rules"][0]["name"] == "*.tgz"
    assert tdata["filter_rules"][0]["type"] == "file"

    tdata.add_filter_rule("tmp")
    assert len(tdata["filter_rules"]) == 2
    assert tdata["filter_rules"][1]["DATA_TYPE"] == "filter_rule"
    assert tdata["filter_rules"][1]["method"] == "exclude"
    assert tdata["filter_rules"][1]["name"] == "tmp"
    assert tdata["filter_rules"][1]["type"] == MISSING


@pytest.mark.parametrize(
    "filter,expected",
    [
        ("foo", "foo"),
        ({"foo": "bar", "a": ["b", "c"]}, "foo:bar/a:b,c"),
        (
            [
                "foo",
                {
                    "a": ["b", "c"],
                },
            ],
            ["foo", "a:b,c"],
        ),
    ],
)
def test_ls_format_filter(filter, expected):
    assert _format_filter(filter) == expected


@pytest.mark.parametrize(
    "deadline_str",
    [
        pytest.param("2026-02-27 09:36:50.517299", id="naive_w_microseconds"),
        pytest.param("2020-01-01 08:22:12+04:00", id="plus4"),
    ],
)
def test_transfer_data_stringifies_deadline(deadline_str):
    deadline_dt = datetime.datetime.fromisoformat(deadline_str)

    tdata = TransferData(GO_EP1_ID, GO_EP2_ID, deadline=deadline_dt)
    assert tdata["deadline"] == deadline_str



import abc

import pytest

from globus_sdk._payload import AbstractGlobusPayload, GlobusPayload


def test_payload_methods():
    # just make sure that PayloadWrapper acts like a dict...
    data = GlobusPayload()
    assert "foo" not in data
    with pytest.raises(KeyError):
        data["foo"]
    data["foo"] = 1
    assert "foo" in data
    assert data["foo"] == 1
    del data["foo"]
    assert "foo" not in data
    assert len(data) == 0
    assert list(data) == []
    data["foo"] = 1
    data["bar"] = 2
    assert len(data) == 2
    assert data == {"foo": 1, "bar": 2}
    data.update({"x": "hello", "y": "world"})
    assert data == {"foo": 1, "bar": 2, "x": "hello", "y": "world"}


def test_abstract_payload_detects_abstract_methods():
    # A has no abstract methods so it will instantiate
    class A(AbstractGlobusPayload):
        pass

    A()

    # B has an abstract method and inherits from AbstractGlobusPayload so it should
    # fail to instantiate
    class B(A):
        @abc.abstractmethod
        def f(self): ...

    with pytest.raises(
        TypeError,
        match=(
            "Can't instantiate abstract class B without an "
            "implementation for abstract method 'f'"
        ),
    ):
        B()

    # C has two abstract methods, so these should be listed comma separated
    class C(B):
        @abc.abstractmethod
        def g(self): ...

    with pytest.raises(
        TypeError,
        match=(
            "Can't instantiate abstract class C without an "
            "implementation for abstract methods ('f', 'g'|'g', 'f')"
        ),
    ):
        C()

    # D should be instantiable because it defines the abstract methods
    class D(C):
        def f(self):
            return 1

        def g(self):
            return 2

    D()



import uuid

from globus_sdk import AuthClient

zero_id = uuid.UUID(int=0)

# ok usages
ac = AuthClient()
ac.get_identities(ids="foo")
ac.get_identities(ids=zero_id)
ac.get_identities(ids=("foo", "bar"))
ac.get_identities(ids=(zero_id,))
ac.get_identities(usernames="foo,bar")
ac.get_identities(usernames=("foo", "bar"))
ac.get_identities(usernames=("foo", "bar"), provision=True)
ac.get_identities(usernames=("foo", "bar"), query_params={"provision": False})

# bad usage
ac.get_identities(usernames=zero_id)  # type: ignore[arg-type]
ac.get_identities(usernames=(zero_id,))  # type: ignore[arg-type]


# test the response object is iterable
res = ac.get_identities(usernames="foo")
for x in res:
    print(x)
